"""
Slot Filling
============

While we describe :ref:`slots <Slot>` in the linked doc. The slot filling is part of an :ref:`intent's<Intent>`
role. This plugin orchestrates the execution of slot filling via methods available on intents.

#. We eagerly apply slot placeholders using the :ref:`apply<ApplySlot>` method. 
#. Once entities are found, we check if the slots can fill them using the :ref:`fill<FillSlot>` method.
#. If no slots were filled, we remove the placeholders using :ref:`cleanup<CleanupSlot>` method.
"""
from typing import List, Optional

from dialogy.base import Guard, Input, Output, Plugin
from dialogy.types import BaseEntity
from dialogy.types.intent import Intent
from dialogy.types.slots import Rule


class RuleBasedSlotFillerPlugin(Plugin):
    """
    A utility :ref:`plugin <AbstractPlugin>` for
    `slot filling <https://nlpprogress.com/english/intent_detection_slot_filling.html>`._
    An :ref:`Intent <intent>` may have a few slots that need to be filled.

    This plugin can assist filling pertaining to certain intent:entity:slot-name association rules.

    :ref:`Schema for rules<ApplySlot>`

    Let's run through a practical example. We will create a workflow and preset the output to have expected intent and
    entities.

    .. _RuleBasedSlotFillerPlugin:

    :param rules: A mapping that defines relationship between an intent, its slots and the entities that fill them.
    :type rules: Rule
    :param fill_multiple: More than one item be allowed to fill a single slot.
    :type fill_multiple: bool
    :param access: Signature for workflow access is :code:`access(workflow) -> (Intent, List[BaseEntity])`
    :type access: Optional[PluginFn]
    :param mutate: Not needed, pass :code:`None`.
    :type mutate: Optional[PluginFn]

    Irrespective of the entities that are found, only the listed type in the slot shall be present in `values`.
    """

    def __init__(
        self,
        rules: Rule,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        replace_output: bool = True,
        fill_multiple: bool = True,
        debug: bool = False,
    ) -> None:
        """
        constructor
        """
        # rules
        #
        # A `Dict` where each key is an intent name, and each value is another `Dict`,
        # in which, each key is an entity and value contains the `slot_name` and `entity_type.`
        #
        # example:
        # ```
        # rules = {"intent": {"slot_name": "entity_type"}}
        # ```
        super().__init__(
            dest=dest, guards=guards, debug=debug, replace_output=replace_output
        )
        self.rules: Rule = rules or {}

        # fill_multiple
        # A boolean value that commands the slot filler to add multiple values of the
        # same entity type within a slot.
        self.fill_multiple = fill_multiple

    def fill(self, intents: List[Intent], entities: List[BaseEntity]) -> List[Intent]:
        if not intents:
            return intents

        intent, *rest = intents

        intent.apply(self.rules)

        for entity in entities:
            intent.fill_slot(entity, fill_multiple=self.fill_multiple)

        intent.cleanup()
        return [intent, *rest]

    def utility(self, _: Input, output: Output) -> List[Intent]:
        return self.fill(output.intents, output.entities)



import os
from datetime import datetime
from typing import Any, Optional


def load_file(
    file_path: Optional[str] = None, mode: str = "r", loader: Any = None
) -> Any:
    """
    Safely load a file.

    :param file_path: The path to the file to load.
    :type file_path: [type]
    :param mode: The mode to use when opening the file., defaults to "r"
    :type mode: str, optional
    :return: The file contents.
    :rtype: Any
    """
    if not isinstance(file_path, str):
        return None

    if os.path.exists(file_path):
        with open(file_path, mode) as file:
            if not loader:
                return file.read()
            else:
                return loader(file)
    else:
        return None


def save_file(
    file_path: Optional[str] = None,
    content: Any = None,
    mode: str = "w",
    encoding: str = "utf-8",
    newline: str = "\n",
    writer: Any = None,
) -> None:
    """
    Save a file.

    :param file_path: The path to the file to save.
    :type file_path: str
    :param content: The content to save.
    :type content: Any
    :param mode: The mode to use when opening the file., defaults to "w"
    :type mode: str, optional
    :param encoding: The encoding to use when writing the file, defaults to "utf-8"
    :type encoding: str, optional
    :param newline: The newline character to use when writing the file, defaults to "\n"
    :type newline: str, optional
    """
    if not isinstance(file_path, str):
        return None

    if mode == "wb":
        with open(file_path, mode) as file:
            _ = file.write(content) if not writer else writer(content, file)
    else:
        with open(file_path, mode, encoding=encoding, newline=newline) as file:
            _ = file.write(content) if not writer else writer(content, file)


def create_timestamps_path(
    directory: str,
    file_name: str,
    timestamp: Optional[datetime] = None,
    dry_run: bool = False,
) -> str:
    timestamp = timestamp or datetime.now()
    dir_path = os.path.join(
        directory, timestamp.strftime("%d-%B-%Y"), timestamp.strftime("%H-%M-%S-%f")
    )
    if not dry_run:
        os.makedirs(dir_path, exist_ok=True)
    return os.path.join(dir_path, file_name)



"""
.. _AbstractPlugin:

Abstract Plugin
===============

A Plugin transforms the data within the workflow. We use plugins for producing :ref:`intents <Intent>`, :ref:`entities <BaseEntity>`,
filling :ref:`slots <Slot>` and even preparing intermediate data for other plugins. This package ships the abstract base class for creating
other plugins. These offer a base for a few already implemented features like :ref:`book-keeping <PluginBookkeeping>` and :ref:`Guards<Guards>`.

Writing Plugins
================

.. attention::

    We will learn plugin creation by exploring some examples. **It is highly advised to read 
    the** :ref:`Input <Input>` and :ref:`Output <Output>` **sections first.**

Keyword Intent
----------------

Say we have to predict an intent *_greeting_* if the first item in the transcripts have :code:`"hello"`.
This is assuming that the transcripts are ranked by their confidence scores in the descending order.

.. ipython ::

    In [1]: from typing import List
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input, Plugin, Output
       ...: from dialogy.types import Intent

    In [2]: class Transcript2IntentPlugin(Plugin):
       ...:     def __init__(self, dest=None, **kwargs):
       ...:         super().__init__(dest=dest, **kwargs)
       ...:         
       ...:     def greet(self, transcripts: List[str]) -> List[Intent]:
       ...:         if not transcripts:
       ...:             return []
       ...:
       ...:         best_transcript, *rest = transcripts
       ...:
       ...:         if "hello" in best_transcript:
       ...:             return [Intent(name="_greeting_", score=1.0)]
       ...:         else:
       ...:             return []
       ...:
       ...:     def utility(self, input_: Input, output: Output) -> List[Intent]:
       ...:         return self.greet(input_.transcripts)

    In [3]: transcript2intent   = Transcript2IntentPlugin(dest="output.intents")
       ...: workflow            = Workflow([transcript2intent])
       ...: input_, output      = workflow.run(Input(utterances=[[{"transcript": "hello"}]]))
       ...: # The last expression provides us a snapshot of the workflow's 
       ...: # input and output after all plugins have been executed.
       ...: # These aren't the Input and Output objects, but their `dict` equivalents.

    In [4]: # we can see the utterance that we set 
       ...: # and the derived field `transcript` that we used within the plugin.
       ...: # Other values are defaults.
       ...: input_

    In [5]: # we can see that the _greeting_ intent has been set.
       ...: output

Heuristic based Intent
------------------------

This problem requires us to predict *_greeting_* if the word :code:`"hello"`
is present at least 6 times across all utterances. To help us with this, 
we will import an existing plugin: :ref:`MergeASROutputPlugin <merge_asr_output_plugin>`.

.. ipython ::
    :okwarning:

    In [6]: from collections import Counter
       ...: from dialogy.plugins import MergeASROutputPlugin

    In [7]: class HeuristicBasedIntentPlugin(Plugin):
       ...:     def __init__(self, threshold=0, dest=None, **kwargs):
       ...:         super().__init__(dest=dest, **kwargs)
       ...:         self.threshold = threshold
       ...: 
       ...:     def greet(self, clf_features: List[str]) -> List[Intent]:
       ...:         if not clf_features:
       ...:             return []
       ...:
       ...:         feature = clf_features[0]
       ...:         word_frequency = Counter(feature.split())
       ...:         if word_frequency.get("hello", 0) >= self.threshold:
       ...:             return [Intent(name="_greeting_", score=1.0)]
       ...:         else:
       ...:             return []
       ...:     def utility(self, input_: Input, output: Output) -> List[Intent]:
       ...:         return self.greet(input_.clf_feature)

    In [8]: heuristic_based_intent  = HeuristicBasedIntentPlugin(threshold=6, dest="output.intents")
       ...: merge_asr_output_plugin = MergeASROutputPlugin(dest="input.clf_feature")
       ...: workflow                = Workflow([merge_asr_output_plugin, heuristic_based_intent])
       ...: input_, output          = workflow.run(
       ...:    Input(utterances=[
       ...:        [{"transcript": "hello is anyone there"},
       ...:         {"transcript": "yellow is anyone here"},
       ...:         {"transcript": "hello is one here"},
       ...:         {"transcript": "hello is one there"},
       ...:         {"transcript": "hello if one here"},
       ...:         {"transcript": "hello in one here"},
       ...:         {"transcript": "hello ip one here"}]
       ...:    ])
       ...: )

    In [9]: # let's check the snapshots again.
       ...: # Pay close attention to `clf_feature`
       ...: # it wasn't set in the previous example.
       ...: # This time, it was set by the `MergeASROutputPlugin` plugin 
       ...: # as it had the dest = "input.clf_feature"
       ...: input_

    In [10]: output


Guarding plugins
----------------

It may seem useful to not run plugins all the time. In this case, if we knew the current state
of the conversation we could decide to not run our naive plugin. Say a state where we are expecting numbers? 
So let's look at an example where we prevent plugins from execution using :code:`guards`.

.. ipython ::
    :okexcept:

    In [1]: heuristic_based_intent  = HeuristicBasedIntentPlugin(
       ...:     threshold=6,
       ...:     dest="output.intents",
       ...:     guards=[lambda i, o: i.current_state == "STATE_EXPECTING_NUMBERS"]
       ...: )
       ...: merge_asr_output_plugin = MergeASROutputPlugin(dest="input.clf_feature")
       ...: workflow                = Workflow([merge_asr_output_plugin, heuristic_based_intent])
       ...: in_                     = Input(utterances=[[
       ...:                                 {"transcript": "hello is anyone there"},
       ...:                                 {"transcript": "yellow is anyone here"},
       ...:                                 {"transcript": "hello is one here"},
       ...:                                 {"transcript": "hello is one there"},
       ...:                                 {"transcript": "hello if one here"},
       ...:                                 {"transcript": "hello in one here"},
       ...:                                 {"transcript": "hello ip one here"}]]) 
       ...: input_, output          = workflow.run(in_)

    In [2]: # Oops we forgot to set the current state within the Input!
       ...: output

    In [3]: workflow            = Workflow([merge_asr_output_plugin, heuristic_based_intent])
       ...: in_.current_state   = "STATE_EXPECTING_NUMBERS"
       ...: # We get a FrozenInstanceError because Input and Output instances are immutable.

    In [4]: # We can use the following to create a new instance of Input instead.
       ...: in_ = Input.from_dict({"current_state": "STATE_EXPECTING_NUMBERS"}, reference=in_)
       ...: input_, output  = workflow.run(in_)

    In [4]: # In this case, we don't see anything set as the 
       ...: # HeuristicBasedIntentPlugin was prevented
       ...: # by its guarding conditions.
       ...: output

    In [5]: # However, `clf_features` are set since we never wrote guards for 
       ...: # the MergeASROutputPlugin.
       ...: input_

Update Plans
------------

You may need to write plugins that generate an :ref:`Intent <Intent>` or :ref:`BaseEntity <BaseEntity>`, there
is also a category of plugins that might be required for modifications. Like the :ref:`RuleBasedSlotFillerPlugin <RuleBasedSlotFillerPlugin>`
or the :ref:`CombineDateTimeOverSlots <CombineDateTimeOverSlots>`.

The former updates the :ref:`intents<Intent>` and the latter updates :ref:`time-entities <TimeEntity>`. In these cases, the plugins
also take into account other values. 

Such as, :ref:`CombineDateTimeOverSlots <CombineDateTimeOverSlots>`: 

#. Separates time entities from other entity types.
#. Combines them over slot presence.
#. Rebuilds a full entity list.

This means

#. Some plugins need to replace the entire set of intents or entities.
#. While others need to append their results along with the history of the workflow.

If your plugin belongs to [1], then you need to set :code:`replace_output=True` in the constructor of your plugin.

.. note:: 

    Refer to :ref:`CombineDateTimeOverSlots <CombineDateTimeOverSlots>`.
    It's :ref:`combine_time_entities_from_slots <combinetimeentitiesfromslots>` method shows where we may need to use
    :code:`replace_output=True`.

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, List, Optional

import dialogy.constants as const
from dialogy.base.input import Input
from dialogy.base.output import Output
from dialogy.utils.logger import logger

if TYPE_CHECKING:  # pragma: no cover
    from dialogy.workflow import Workflow


Guard = Callable[[Input, Output], bool]


class Plugin(ABC):
    """
    Abstract class to be implemented by all plugins.

    :param input_column: Transforms data in this column for a given dataframe, defaults to const.ALTERNATIVES
    :type input_column: str
    :param output_column: Saves transformation in this column for a given dataframe, defaults to None
    :type output_column: Optional[str]
    :param use_transform: Should the transformation be applied while training?, defaults to False
    :type use_transform: bool
    :param dest: The path where plugin output should be saved., defaults to None
    :type dest: Optional[str]
    :param guards: A list of functions that evaluate to bool, defaults to None
    :type guards: Optional[List[Guard]]
    :param debug: Should we print debug logs?, defaults to False
    :type debug: bool, optional
    """

    def __init__(
        self,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = False,
        replace_output: bool = False,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        debug: bool = False,
    ) -> None:
        self.debug = debug
        self.guards = guards
        self.dest = dest
        self.replace_output = replace_output
        self.input_column = input_column
        self.output_column = output_column or input_column
        self.use_transform = use_transform

    @abstractmethod
    def utility(self, input_: Input, output: Output) -> Any:
        """
        An abstract method that describes the plugin's functionality.

        :param input: The workflow's input.
        :type input: Input
        :param output: The workflow's output.
        :type output: Output
        :return: The value returned by the plugin.
        :rtype: Any
        """

    def __call__(self, workflow: Workflow) -> None:
        """
        Set workflow with output.

        This important book-keeping method is called by the workflow.

        - If the plugin guards evaluate to :code:`True`, we don't run the plugin's business logic.
        - Otherwise, we obtain the plugins transformation and set it on :code:`self.dest` path within the workflow.
            - This path is a string describing :code:`'input'` or :code:`'output'` and their respective attributes separated by a '.'
            - like: :code:`"input.transcripts"` or :code:`"output.intents"`.
        - The workflow takes care of keeping its :ref:`input<Input>` and :ref:`output<Output>` immutable.

        .. _PluginBookkeeping:

        :param workflow: An instance of :ref:`Workflow <WorkflowClass>`.
        :type workflow: Workflow
        """
        logger.enable(str(self)) if self.debug else logger.disable(str(self))

        if workflow.input is None:
            return

        if workflow.output is None:
            return

        if self.prevent(workflow.input, workflow.output):
            return

        value = self.utility(workflow.input, workflow.output)
        if value is not None and isinstance(self.dest, str):
            workflow.set(self.dest, value, self.replace_output)

    def prevent(self, input_: Input, output: Output) -> bool:
        """
        Decide if the plugin should execute.

        If this method returns true, the plugin's utility method will not be called.

        .. _Guards:

        :return: prevent plugin execution if True.
        :rtype: bool
        """
        if not self.guards:
            return False
        return any(guard(input_, output) for guard in self.guards)

    def train(self, _: Any) -> Any:
        """
        Train a plugin.
        """
        return None

    def transform(self, training_data: Any) -> Any:
        """
        Transform data for a plugin in the workflow.
        """
        if not self.use_transform:
            return training_data
        return training_data

    def __str__(self) -> str:
        return self.__class__.__name__



"""
.. _CreditCardNumberEntity

Credit Card Number Entity
=========================

Provides the entity class for credit card numbers in natural language. This entity is obtained via :ref:`DucklingPlugin<DucklingPlugin>`.

Plugin Walkthrough
------------------

.. ipython::

    In [1]: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["credit-card-number"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: duckling_plugin.parse("my credit card number is 4111111111111111.")

Workflow Integration
--------------------

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.workflow import Workflow

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["credit-card-number"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: _, output = workflow.run(Input(utterances="my credit card number is 4111111111111111."))

    In [5]: output

"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

import attr
from this import d

from dialogy import constants as const
from dialogy.types import BaseEntity


@attr.s
class CreditCardNumberEntity(BaseEntity):
    entity_type: Optional[str] = attr.ib(
        repr=False, default="credit-card-number", kw_only=True
    )
    issuer: str = attr.ib(validator=attr.validators.instance_of(str), kw_only=True)
    value: Optional[str] = attr.ib(default=None, kw_only=True)
    values: List[Dict[str, Any]] = attr.ib(
        validator=attr.validators.instance_of(list), kw_only=True
    )

    @classmethod
    def from_duckling(
        cls, d: Dict[str, Any], alternative_index: int
    ) -> CreditCardNumberEntity:
        value = d[const.VALUE][const.VALUE]
        return cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            latent=d[const.LATENT],
            values=[{const.VALUE: value}],
            issuer=d[const.VALUE][const.ISSUER],
        )



import copy
import json
import re
import traceback
from typing import Any, Callable, List, Optional

import pandas as pd
from loguru import logger
from tqdm import tqdm

import dialogy.constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.types import BaseEntity
from dialogy.utils import normalize


def get_entity_type(entity: BaseEntity) -> str:
    return f"<{entity.entity_type}>"


class CanonicalizationPlugin(Plugin):
    """
    This plugin implements the canonicalization of the text.
    """

    def __init__(
        self,
        serializer: Callable[[BaseEntity], str] = get_entity_type,
        mask: str = "MASK",
        mask_tokens: Optional[List[str]] = None,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        entity_column: str = const.ENTITY_COLUMN,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = False,
        threshold: float = 0.0,
        debug: bool = False,
    ) -> None:
        super().__init__(
            dest=dest,
            guards=guards,
            debug=debug,
            use_transform=use_transform,
            input_column=input_column,
            output_column=output_column,
        )
        self.mask = mask
        self.mask_tokens = mask_tokens or []
        self._mask_patterns: List[re.Pattern[str]] = [
            re.compile(r"\b%s\b" % token, re.I | re.U) for token in self.mask_tokens
        ]
        self.serializer = serializer
        self.threshold = threshold
        self.entity_column = entity_column

    def mask_transcript(
        self, entities: List[BaseEntity], transcripts: List[str]
    ) -> List[str]:
        canonicalized_transcripts = copy.copy(transcripts)
        for entity in entities:
            if entity.score is not None and entity.score < self.threshold:
                continue
            if entity.alternative_index is None or not entity.alternative_indices:
                entity  # pylint: disable=pointless-statement
                continue
            start = entity.range[const.START]
            end = entity.range[const.END]
            transcript = transcripts[entity.alternative_index]
            canon = f"{transcript[:start]}{get_entity_type(entity)}{transcript[end:]}"
            for idx in entity.alternative_indices:
                canonicalized_transcripts[idx] = canon

        for i, transcript in enumerate(canonicalized_transcripts):
            for pattern in self._mask_patterns:
                canonicalized_transcripts[i] = pattern.sub(self.mask, transcript)

        return canonicalized_transcripts

    def utility(self, input: Input, output: Output) -> Any:
        entities = output.entities
        transcripts = input.transcripts
        return self.mask_transcript(entities, transcripts)

    def transform(self, training_data: pd.DataFrame) -> pd.DataFrame:
        if not self.use_transform:
            return training_data

        logger.debug(f"Transforming dataset via {self.__class__.__name__}")

        for i, row in tqdm(training_data.iterrows(), total=len(training_data)):
            try:
                canonicalized_transcripts = self.mask_transcript(
                    row[self.entity_column],
                    normalize(json.loads(row[self.input_column])),
                )
                training_data.loc[i, self.output_column] = json.dumps(
                    canonicalized_transcripts
                )
            except Exception as error:  # pylint: disable=broad-except
                logger.error(
                    f"{error} -- {row[self.input_column]}\n{traceback.format_exc()}"
                )
        return training_data



"""
.. _DurationEntity:

Duration Entity
================

Provides the entity class for durations in natural language. This entity is obtained via :ref:`DucklingPlugin<DucklingPlugin>`.

Plugin Walkthrough
------------------

.. ipython::

    In [1]: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["duration"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: duckling_plugin.parse("she said 2h.")

Workflow Integration
--------------------

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.workflow import Workflow

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["duration"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: _, output = workflow.run(Input(utterances="she said 2h"))

    In [5]: output
"""
from __future__ import annotations

import operator
from datetime import timedelta
from typing import Any, Dict

import attr

from dialogy import constants as const
from dialogy.types.entity.base_entity import BaseEntity
from dialogy.types.entity.time import TimeEntity
from dialogy.utils import unix_ts_to_datetime


@attr.s
class DurationEntity(BaseEntity):
    """
    This entity type expects a normalized attribute. This provides the duration normalized to seconds.

    Helpful in cases where we wish to operate on time like:
    "I want a booking in 2 days."

    We can tell the time at which the sentence was said, but we need to make the booking after two days.

    This entity parses this information and also provides us the number of seconds to add to the current timestamp
    to get to a date that's 2 days ahead.
    """

    unit: str = attr.ib(validator=attr.validators.instance_of(str), kw_only=True)
    normalized: Dict[str, Any] = attr.ib(default=attr.Factory(dict))
    _meta: Dict[str, str] = attr.ib(default=attr.Factory(dict))
    entity_type: str = attr.ib(default="duration", kw_only=True)

    @classmethod
    def from_duckling(cls, d: Dict[str, Any], alternative_index: int) -> DurationEntity:
        value = d[const.VALUE][const.NORMALIZED][const.VALUE]
        return cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            latent=d[const.LATENT],
            values=[{const.VALUE: value}],
            unit=d[const.VALUE][const.UNIT],
            normalized=d[const.VALUE][const.NORMALIZED],
        )

    def as_time(
        self, reference_unix_ts: int, timezone: str, duration_cast_operator: str
    ) -> TimeEntity:
        """
        Converts a duration entity to a time entity.
        """
        reference_datetime = unix_ts_to_datetime(reference_unix_ts, timezone=timezone)
        if duration_cast_operator == const.FUTURE:
            operation = operator.add
        elif duration_cast_operator == const.PAST:
            operation = operator.sub

        value_dt = operation(
            reference_datetime, timedelta(seconds=self.normalized[const.VALUE])
        )
        value = value_dt.isoformat()

        entity = TimeEntity(
            range={
                const.START: self.range[const.START],
                const.END: self.range[const.END],
            },
            score=self.score,
            body=self.body,
            dim="time",
            alternative_index=self.alternative_index,
            latent=self.latent,
            values=[{const.VALUE: value}],
            grain=self.unit,
        )
        entity.set_entity_type()
        return entity



"""
.. _currency_entity:
Module provides access to entity types that can be parsed to currencies and their value.

Import classes:
    - CurrencyEntity
"""
from __future__ import annotations

from typing import Any, Dict, Optional

import attr

from dialogy import constants as const
from dialogy.types.entity.base_entity import BaseEntity


@attr.s
class CurrencyEntity(BaseEntity):
    """
    Numerical Entity Type

    Use this type for handling all entities that can be parsed to obtain:
    - numbers
    - date
    - time
    - datetime

    Attributes:
        - `dim` dimension of the entity from duckling parser
        - `type` is the type of the entity which can have values in ["value", "interval"]
    """

    unit: str = attr.ib(validator=attr.validators.instance_of(str), kw_only=True)
    entity_type: str = attr.ib(
        default="amount-of-money",
        validator=attr.validators.instance_of(str),
        kw_only=True,
    )

    def get_value(self) -> Any:
        """
        Getter for CurrencyEntity.

        We are yet to decide the pros and cons of the output. It seems retaining {"value": float, "unit": }

        :param reference: [description], defaults to None
        :type reference: Any, optional
        :return: [description]
        :rtype: Any
        """
        value = super().get_value()
        return f"{self.unit}{value:.2f}"

    @classmethod
    def from_duckling(cls, d: Dict[str, Any], alternative_index: int) -> CurrencyEntity:
        value = d[const.VALUE][const.VALUE]
        return cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            values=[{const.VALUE: value}],
            latent=d[const.LATENT],
            unit=d[const.VALUE][const.UNIT],
        )



from dialogy.utils.datetime import (
    dt2timestamp,
    is_unix_ts,
    make_unix_ts,
    unix_ts_to_datetime,
)
from dialogy.utils.file_handler import create_timestamps_path, load_file, save_file
from dialogy.utils.logger import logger
from dialogy.utils.misc import traverse_dict, validate_type
from dialogy.utils.naive_lang_detect import lang_detect_from_text
from dialogy.utils.normalize_utterance import is_utterance, normalize



"""
.. _DucklingPlugin:

We use `Duckling <https://github.com/facebook/duckling>`_ for parsing values like: :code:`date`, :code:`time`,
:code:`numbers`, :code:`currency` etc from natural language. The parser will expect Duckling to be running as an http service, and provide
means to connect from the implementation here. Here's a big example for various duckling entities.

Mother of all examples
----------------------

.. ipython::

    In [1]: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=[
       ...:         "number",
       ...:         "people",
       ...:         "time",
       ...:         "duration",
       ...:         "intersect",
       ...:         "amount-of-money",
       ...:         "credit-card-number",
       ...:     ],
       ...:     # Duckling supports multiple dimensions, by specifying a list, we make sure
       ...:     # we are searching for a match within the expected dimensions only.
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: utterances = [
       ...:    "between today 7pm and tomorrow 9pm",
       ...:    "can we get up at 6 am",
       ...:    "we are 7 people",
       ...:    "2 children 1 man 3 girls",
       ...:    "can I come now?",
       ...:    "can I come tomorrow",
       ...:    "how about monday then",
       ...:    "call me on 5th march",
       ...:    "I want 4 pizzas",
       ...:    "2 hours",
       ...:    "can I pay $5 instead?",
       ...:    "my credit card number is 4111111111111111",
       ...:]

    In [5]: %%time
       ...: input_, output = workflow.run(Input(utterances=utterances))

    In [6]: output


Testing
-------

1. Connect to the `Duckling API <https://github.com/facebook/duckling>`_ either via a docker container or a local setup.
2. Boot an IPython session and setup an instance of :ref:`DucklingPlugin <DucklingPlugin>`.

.. code-block:: python
    :linenos:

    from pprint import pprint
    from dialogy.workflow import Workflow
    from dialogy.base import Input
    from dialogy.plugins import DucklingPlugin

    duckling_plugin = DucklingPlugin(
        dest="output.entities",
        dimensions=["people", "time"],
        locale="en_IN",
        timezone="Asia/Kolkata",
        origin/master
    )

    entities = duckling_plugin.parse(["We are 2 children coming tomorrow at 5 pm"])
    pprint(entities)
    # [PeopleEntity(
    #     body='2 children', 
    #     type='value', 
    #     parsers=['DucklingPlugin'], 
    #     score=1.0, 
    #     alternative_index=0,
    #     alternative_indices=[0, 0], 
    #     latent=False, 
    #     value=2, 
    #     origin='value', 
    #     unit='child',
    #     entity_type='people')
    # ,TimeEntity(
    #     body='tomorrow at 5 pm', 
    #     type='value', 
    #     parsers=['DucklingPlugin'], 
    #     score=1.0, 
    #     alternative_index=0, 
    #     alternative_indices=[0],
    #     latent=False, 
    #     value='2022-02-11T17:00:00.000+05:30', origin='value', entity_type='datetime', grain='hour')]

    # To convert these to dicts:

    entity_dicts = [entity.json() for entity in entities]
    pprint(entity_dicts)
    # [{'range': {'start': 7, 'end': 17},
    #     'body': '2 children',
    #     'type': 'value',
    #     'parsers': ['DucklingPlugin'],
    #     'score': 1.0,
    #     'alternative_index': 0,
    #     'value': 2,
    #     'unit': 'child',
    #     'entity_type': 'people'},
    # {'range': {'start': 38, 'end': 54},
    #     'body': 'tomorrow at 5 pm',
    #     'type': 'value',
    #     'parsers': ['DucklingPlugin'],
    #     'score': 1.0,
    #     'alternative_index': 0,
    #     'value': '2022-02-11T17:00:00.000+05:30',
    #     'entity_type': 'datetime',
    #     'grain': 'hour'
    # }]

Guards
------

Like all plugins, the Duckling plugin can be :ref:`guarded <Guards>` by conditions that prevent its execution.
You may need this if you want to save on latency and are sure of not expecting entities in those turns. We will use
the :ref:`mother of all examples <Mother of all examples>` from above to make the point. We will use the :code:`current_state` to tell the plugin
to guard itself in the *"SMALL_TALK"* state.

.. ipython::

    In [1]: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=[
       ...:         "number",
       ...:         "people",
       ...:         "date",
       ...:         "time",
       ...:         "duration",
       ...:         "intersect",
       ...:         "amount-of-money",
       ...:         "credit-card-number",
       ...:     ],
       ...:     guards=[lambda i, o: i.current_state == "SMALL_TALK"]
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: utterances = [
       ...:    "between today 7pm and tomorrow 9pm",
       ...:    "can we get up at 6 am",
       ...:    "we are 7 people",
       ...:    "2 children 1 man 3 girls",
       ...:    "can I come now?",
       ...:    "can I come tomorrow",
       ...:    "how about monday then",
       ...:    "call me on 5th march",
       ...:    "I want 4 pizzas",
       ...:    "2 hours",
       ...:    "can I pay $5 instead?",
       ...:    "my credit card number is 4111111111111111",
       ...:]

    In [5]: %%time
       ...: input_, output = workflow.run(Input(utterances=utterances, current_state="SMALL_TALK"))

    In [6]: output

You can notice that the time drops from **ms** to **us** and we produce no entities.

Filtering Date and Time
-----------------------

There are cases where we run into ambiguity in entity resolution. A fun case happens usually with
the Hindi language when someone says *"कल"*. The problem is, *"कल"* could be *"tommorrow"* or *"yesterday"*.

Problem
^^^^^^^

.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: utterances = ["कल"]

    In [5]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)

    In [5]: input_, output = workflow.run(Input(utterances=utterances, locale="hi_IN", reference_time=reference_time))

    In [6]: output

See the two entities? 

- We froze our reference time for 1st Jan 2022 at 12:00:00. 
- We get 31st December 2021
- and 2nd Jan 2022.

This could be decoded using additional context. Let's assume it is a given that dates can only be in future. Say
a flight-booking case.

.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...:     datetime_filters=DucklingPlugin.FUTURE
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: utterances = ["कल"]

    In [5]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)

    In [5]: input_, output = workflow.run(Input(utterances=utterances, locale="hi_IN", reference_time=reference_time))

    In [6]: output

We can use :code:`DucklingPlugin.PAST` for inverse. 

What if the context was within the utterance? Say for payment verification cases, we receive these utterances:

- *"कल ही कर दी थी"* -- *"I have already paid yesterday"*
- *"कल कर दूंगा"* -- *"I will pay tomorrow"*

We can't filter out either future or past values. In these cases, the `intent <Intents>` carries the context
for resolving the entity value. We will look into :ref:`Temporal Intents <Temporal Intents>` for more details.


Temporal Intents
-----------------

Filtering
^^^^^^^^^

There are cases where intents carry a notion of time, repeating an example from above:

#. *"कल ही कर दी थी"* -- *"I have already paid yesterday"*
#. *"कल कर दूंगा"* -- *"I will pay tomorrow"*

While we can't decode the entity value for "कल" in the above utterances, the correct intent prediction is:

#. :code:`already_paid`
#. :code:`pay_later`

If we could use this information, we could resolve the entity value for "कल". We will simulate the output
to contain appropriate intents using the :code:`set` method of :ref:`workflow <Workflow>`.

.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.types import Intent
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...:     temporal_intents={
       ...:         "already_paid": DucklingPlugin.FUTURE,
       ...:         "pay_later": DucklingPlugin.PAST
       ...:     }
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: utterances = ["कल ही कर दी थी"]
       ...: workflow.set("output.intents", [Intent(name="already_paid", score=1.0)])

    In [5]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)
       ...: input_, output = workflow.run(Input(utterances=utterances, locale="hi_IN", reference_time=reference_time))

    In [6]: output

    In [7]: utterances = ["कल ही कर दी थी"]
       ...: workflow.set("output.intents", [Intent(name="pay_later", score=1.0)])
       ...: input_, output = workflow.run(Input(utterances=utterances, locale="hi_IN", reference_time=reference_time))

    In [6]: output

Casting
^^^^^^^

We can also use temporal intents to cast certain entities. Machines can act on absolute time values but 
natural language often has relative units like *"for 2 hours"*. These produce a duration entity.

.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["duration", "time"],
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: utterances = ["for 2h"]

    In [5]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)

    In [5]: input_, output = workflow.run(Input(utterances=utterances, reference_time=reference_time))

    In [6]: output

We understand the duration but we could create an absolute time value using the reference time, but we don't know if duration 
should be added or removed.

.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.types import Intent
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time", "duration"],
       ...:     timezone="Asia/Kolkata",
       ...:     temporal_intents={
       ...:         "already_paid": DucklingPlugin.FUTURE,
       ...:         "pay_later": DucklingPlugin.PAST
       ...:     }
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)
       ...: utterances = ["for 2h"]

    In [5]: workflow.set("output.intents", [Intent(name="already_paid", score=1.0)])
       ...: input_, output = workflow.run(Input(utterances=utterances, reference_time=reference_time))

    In [6]: output

    In [7]: workflow.set("output.intents", [Intent(name="pay_later", score=1.0)])
       ...: input_, output = workflow.run(Input(utterances=utterances, reference_time=reference_time))

    In [8]: output

In case we need to always cast duration as a future or a past value we can do:


.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.types import Intent
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time", "duration"],
       ...:     timezone="Asia/Kolkata",
       ...:     temporal_intents={
       ...:         "__any__": DucklingPlugin.FUTURE,
       ...:     }
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)
       ...: utterances = ["for 2h"]

    In [5]: input_, output = workflow.run(Input(utterances=utterances, reference_time=reference_time))

    In [6]: output

    In [7]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time", "duration"],
       ...:     timezone="Asia/Kolkata",
       ...:     temporal_intents={
       ...:         "__any__": DucklingPlugin.PAST,
       ...:     }
       ...: )

    In [8]: workflow = Workflow([duckling_plugin])
       ...: input_, output = workflow.run(Input(utterances=utterances, reference_time=reference_time))
       ...: output

"""
import json
import operator
import traceback
from concurrent import futures
from datetime import datetime
from pprint import pformat
from typing import Any, Callable, Dict, List, Optional, Union

import pandas as pd
import pydash as py_
import pytz
import requests
from pytz.tzinfo import BaseTzInfo
from tqdm import tqdm

from dialogy import constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.base.entity_extractor import EntityScoringMixin
from dialogy.types import BaseEntity, Intent
from dialogy.types.entity.deserialize import deserialize_duckling_entity
from dialogy.utils import dt2timestamp, lang_detect_from_text, logger


class DucklingPlugin(EntityScoringMixin, Plugin):
    """
    :param dimensions: `Dimensions <https://github.com/facebook/duckling#supported-dimensions>`_. Of the listed
    dimensions, we support:

        - `Numeral`
        - `Time`
        - `TimeInterval`
        - `Duration`
        - `People` - This isn't part of the standard, we have a private fork to support this.

        Do note, passing more dimensions is not free. Duckling would search for extra set of patterns just because
        those dimensions were expected.
    :type dimensions: Optional[List[str]]

    :param locale: The format for expressing locale requires language code and country name ids. Read about
        `sections <https://github.com/facebook/duckling#extending-duckling>`_ that define
        `ISO-639-codes <https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes>`_ for languages and
        `ISO3166 alpha2 country code <https://www.iso.org/obp/ui/#search/code/>`_ for country codes.
        Examples: `"en_IN"`, `"en_US"`, `"en_GB"`.
    :type locale: str

    :param timezone: `pytz` Timezone. This is especially important when services are deployed across different
    geographies and consistency is expected in the responses. Get a valid value from a `list of tz database timezones
    <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_. Example: `"Asia/Kolkata"` :type timezone:
    Optional[str]

    :param timeout: There are certain strings which tend to stall Duckling:
    `example <https://github.com/facebook/duckling/issues/338>`_.
    In such cases, to prevent the overall experience to slow
    down as well, provide a certain timeout value, defaults to 0.5.
    :type timeout: float

    :param url: The address where Duckling's entity parser can be reached, defaults to "http://0.0.0.0:8000/parse".
    :type url: Optional[str]
    """

    FUTURE = const.FUTURE
    """
    Select time entity values with scores greater than or equal to the reference time.
    """

    PAST = const.PAST
    """
    Select time entity values with scores lesser than or equal to the reference time.
    """

    __DATETIME_OPERATION_ALIAS = {FUTURE: operator.ge, PAST: operator.le}

    def __init__(
        self,
        dimensions: List[str],
        timezone: str = "UTC",
        timeout: float = 0.5,
        url: str = "http://0.0.0.0:8000/parse",
        locale: str = "en_IN",
        temporal_intents: Optional[Dict[str, str]] = None,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        datetime_filters: Optional[str] = None,
        threshold: Optional[float] = None,
        activate_latent_entities: Union[Callable[..., bool], bool] = False,
        reference_time_column: str = const.REFERENCE_TIME,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = False,
        debug: bool = False,
    ) -> None:
        """
        constructor
        """
        super().__init__(
            dest=dest,
            guards=guards,
            debug=debug,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
        )
        self.dimensions = dimensions
        self.locale = locale
        self.timezone = timezone
        self.timeout = timeout
        self.threshold = threshold
        self.url = url
        self.temporal_intents = temporal_intents or {}
        self.reference_time: Optional[int] = None
        self.reference_time_column = reference_time_column
        self.datetime_filters = datetime_filters
        self.activate_latent_entities = activate_latent_entities
        self.session = requests.Session()
        self.session.mount(
            "http://",
            requests.adapters.HTTPAdapter(
                max_retries=1, pool_maxsize=10, pool_block=True
            ),
        )
        self.headers: Dict[str, str] = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8"
        }

    def __set_timezone(self) -> Optional[BaseTzInfo]:
        """
        Set timezone as BaseTzInfo from compatible timezone string.

        Raises:
            pytz.UnknownTimeZoneError:

        :raises pytz.UnknownTimeZoneError: If `self.timezone` is not in `list of tz database
            timezones <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_.
        :return: A valid timezone
        :rtype: Optional[BaseTzInfo]
        """
        # If timezone is an unsafe string, we will handle a `pytz.UnknownTimeZoneError` exception
        # and pass a friendly message.
        try:
            return pytz.timezone(self.timezone)
        except pytz.UnknownTimeZoneError as unknown_timezone_error:
            raise pytz.UnknownTimeZoneError(
                f"The timezone {self.timezone} is invalid"
                " check valid types here: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
            ) from unknown_timezone_error

    def __create_req_body(
        self,
        text: str,
        reference_time: Optional[int] = None,
        locale: str = "en_IN",
        use_latent: Union[Callable[..., bool], bool] = False,
    ) -> Dict[str, Any]:
        """
        create request body for entity parsing

        example: "3 people tomorrow"

        Make your own reference time using the current timestamp using: :code:`int(datetime.now().timestamp() * 1000)`
        These are the milliseconds since the `Unix epoch <https://en.wikipedia.org/wiki/Unix_time>`_

        :param text: A sentence or document.
        :type text: str
        :param reference_time: Impart context of timestamp,
        relevant for time related entities. Resolve relative time like "yesterday", "next month", etc.
        :type Optional[int]
        :param reference_time: Optional[int]
        :return: request object for Duckling API.
        :rtype: Dict[str, Any]
        """
        dimensions = self.dimensions
        self.activate_latent_entities = use_latent or self.activate_latent_entities
        activate_latent_entities = (
            self.activate_latent_entities
            if isinstance(self.activate_latent_entities, bool)
            else self.activate_latent_entities()
        )

        payload = {
            "text": text,
            "locale": locale or self.locale,
            "tz": self.__set_timezone(),
            "dims": json.dumps(dimensions),
            "reftime": reference_time,
            "latent": activate_latent_entities,
        }
        logger.debug("Duckling API payload:")
        logger.debug(pformat(payload))

        return payload

    def get_operator(self, filter_type: Any) -> Any:
        try:
            return getattr(operator, filter_type)
        except (AttributeError, TypeError) as exception:
            logger.debug(traceback.format_exc())
            raise ValueError(
                f"Expected datetime_filters to be one of {self.FUTURE}, {self.PAST} "
                "or a valid comparison operator here: https://docs.python.org/3/library/operator.html"
            ) from exception

    def select_datetime(
        self, entities: List[BaseEntity], filter_type: Any
    ) -> List[BaseEntity]:
        """
        Select datetime entities as per the filters provided in the configuration.

        :param entities: A list of entities.
        :type entities: List[BaseEntity]
        :param filter_type:
        :type filter_type: str
        :return: List of entities obtained after applying comparator functions.
        :rtype: List[BaseEntity]
        """
        if not isinstance(filter_type, str):
            raise TypeError(
                f"Expected datetime_filters to be a str and one of {self.FUTURE}, {self.PAST} or"
                " a valid comparison operator here: https://docs.python.org/3/library/operator.html"
            )

        if filter_type in self.__DATETIME_OPERATION_ALIAS:
            operation = self.__DATETIME_OPERATION_ALIAS[filter_type]
        else:
            operation = self.get_operator(filter_type)

        time_entities, other_entities = py_.partition(
            entities, lambda entity: entity.dim == const.TIME
        )

        filtered_time_entities = [
            entity
            for entity in time_entities
            if operation(dt2timestamp(entity.get_value()), self.reference_time)
        ]
        return filtered_time_entities + other_entities

    def apply_filters(self, entities: List[BaseEntity]) -> List[BaseEntity]:
        """
        Conditionally remove entities.

        The utility of this method is tracked here:
        https://github.com/Vernacular-ai/dialogy/issues/42

        We needed a way to express, not all datetime entities are needed. There are needs which can be
        expressed as filters: want greater or lesser than the reference time, etc. There are more applications
        where we expect sorting, filtering by other attributes as well. This method is an advent of such expressions.

        :param entities: A list of entities.
        :type entities: List[BaseEntity]
        :return: A list of entities obtained after applying filters.
        :rtype: List[BaseEntity]
        """
        if self.datetime_filters:
            return self.select_datetime(entities, self.datetime_filters)

        # We call the filters that exist on the EntityExtractor class like threshold filtering.
        return super().apply_filters(entities)

    def _reshape(
        self,
        entities_json: List[Dict[str, Any]],
        alternative_index: int = 0,
        reference_time: Optional[int] = None,
        duration_cast_operator: Optional[str] = None,
    ) -> List[BaseEntity]:
        """
        Create a list of :ref:`BaseEntity <base_entity>` objects from a list of entity dicts.

        :param entities_json: List of entities derived from Duckling's API.
        :type entities_json: List[Dict[str, Any]]
        :raises NotImplementedError: Raised when dimensions not supported by the project are used.
        :raises KeyError: Expected keys in entity dict don't match the Entity class.
        :return: A list of objects subclassed from :ref:`BaseEntity <base_entity>`
        :rtype: Optional[List[BaseEntity]]
        """
        deserialized_entities = []
        for entity_json in entities_json:
            entity = deserialize_duckling_entity(
                entity_json,
                alternative_index,
                reference_time=reference_time,
                timezone=self.timezone,
                duration_cast_operator=duration_cast_operator,
            )
            entity.add_parser(self)
            deserialized_entities.append(entity)
        return deserialized_entities

    def _get_entities(
        self,
        text: str,
        locale: str = "en_IN",
        reference_time: Optional[int] = None,
        use_latent: Union[Callable[..., bool], bool] = False,
        sort_idx: int = 0,
    ) -> Dict[str, Any]:
        """
        Get entities from duckling-server.

        Assuming duckling-server is running at expected `url`. The entities are returned in
        `json` compatible format.

        :param text: The sentence or document in which entities must be looked up.
        :type text: str
        :param reference_time: Cases where relative units of time are mentioned,
                                like "today", "now", etc. We need to know the current time
                                to parse the values into usable dates/times, defaults to None
        :type reference_time: Optional[int], optional
        :raises ValueError: Duckling API call failure leading to no json response.
        :return: Duckling entities as :code:`dicts`.
        :rtype: List[Dict[str, Any]]
        """
        body = self.__create_req_body(
            text, reference_time=reference_time, locale=locale, use_latent=use_latent
        )

        try:
            response = self.session.post(
                self.url, data=body, headers=self.headers, timeout=self.timeout
            )

            if response.status_code == 200:
                # The API call was successful, expect the following to contain entities.
                # A list of dicts or an empty list.
                return {const.IDX: sort_idx, const.VALUE: response.json()}
        except requests.exceptions.Timeout as timeout_exception:
            logger.error(f"Duckling timed out: {timeout_exception}")  # pragma: no cover
            logger.error(pformat(body))  # pragma: no cover
            return {const.IDX: sort_idx, const.VALUE: []}  # pragma: no cover
        except requests.exceptions.ConnectionError as connection_error:
            logger.error(f"Duckling server is turned off?: {connection_error}")
            logger.error(pformat(body))
            raise requests.exceptions.ConnectionError from connection_error

        # Control flow reaching here would mean the API call wasn't successful.
        # To prevent rest of the things from crashing, we will raise an exception.
        raise ValueError(
            f"Duckling API call failed | [{response.status_code}]: {response.text}"
        )

    def _get_entities_concurrent(
        self,
        texts: List[str],
        locale: str = "en_IN",
        reference_time: Optional[int] = None,
        use_latent: Union[Callable[..., bool], bool] = False,
    ) -> List[List[Dict[str, Any]]]:
        """
        Make multiple-parallel API calls to duckling-server .

        :param texts: A list of strings to be parsed concurrently via duckling.
        :type texts: List[str]
        :param locale: Defined in __init__, defaults to "en_IN"
        :type locale: str, optional
        :param reference_time: Cases where relative units of time are mentioned,
                                like "today", "now", etc. We need to know the current time
                                to parse the values into usable dates/times, defaults to None
        :type reference_time: Optional[int], optional
        :param use_latent: True returns latent entities, defaults to False
        :type use_latent: bool, optional
        :return: Duckling entities as :code:`dicts`.
        :rtype: List[List[Dict[str, Any]]]
        """
        workers = min(10, max(len(texts), 1))
        with futures.ThreadPoolExecutor(max_workers=workers) as executor:
            futures_ = [
                executor.submit(
                    self._get_entities,
                    text,
                    locale,
                    reference_time=reference_time,
                    use_latent=use_latent,
                    sort_idx=i,
                )
                for i, text in enumerate(texts)
            ]
        entities_list = [future.result() for future in futures.as_completed(futures_)]
        return [
            entities[const.VALUE]
            for entities in sorted(
                entities_list, key=lambda entities: entities[const.IDX]
            )
        ]

    def apply_entity_classes(
        self,
        list_of_entities: List[List[Dict[str, Any]]],
        reference_time: Optional[int] = None,
        timezone: str = "UTC",
        duration_cast_operator: Optional[str] = None,
    ) -> List[BaseEntity]:
        shaped_entities = []
        for (alternative_index, entities) in enumerate(list_of_entities):
            shaped_entities.append(
                self._reshape(
                    entities,
                    alternative_index=alternative_index,
                    reference_time=reference_time,
                    duration_cast_operator=duration_cast_operator,
                )
            )
        return py_.flatten(shaped_entities)

    def validate(
        self, input_: Union[str, List[str]], reference_time: Optional[int]
    ) -> "DucklingPlugin":
        input_is_str = isinstance(input_, str)
        inputs_are_list_of_strings = isinstance(input_, list) and all(
            isinstance(text, str) for text in input_
        )
        if not isinstance(reference_time, int) and self.datetime_filters:
            raise TypeError(
                "Duckling requires reference_time to be a unix timestamp (int) but"
                f" {type(reference_time)} was found"
                "https://stackoverflow.com/questions/20822821/what-is-a-unix-timestamp-and-why-use-it\n"
            )

        if not input_is_str and not inputs_are_list_of_strings:
            raise TypeError(f"Expected {input_} to be a List[str] or str.")
        return self

    def parse(
        self,
        transcripts: Union[str, List[str]],
        locale: Optional[str] = None,
        reference_time: Optional[int] = None,
        use_latent: Union[Callable[..., bool], bool] = False,
        intents: Optional[List[Intent]] = None,
    ) -> List[BaseEntity]:
        list_of_entities: List[List[Dict[str, Any]]] = []
        entities: List[BaseEntity] = []
        duration_cast_operator = None
        intent = None

        if intents:
            intent, *_ = intents

        if isinstance(intent, Intent) and intent.name in self.temporal_intents:
            duration_cast_operator = self.datetime_filters = self.temporal_intents.get(
                intent.name
            )

        if const.ANY in self.temporal_intents:
            duration_cast_operator = self.temporal_intents.get(const.ANY)

        locale = locale or self.locale
        self.reference_time = reference_time = reference_time or self.reference_time
        self.validate(transcripts, reference_time)
        if isinstance(transcripts, str):
            transcripts = [transcripts]  # pragma: no cover

        list_of_entities = self._get_entities_concurrent(
            transcripts,
            locale=locale,
            reference_time=reference_time,
            use_latent=use_latent,
        )
        entities = self.apply_entity_classes(
            list_of_entities,
            reference_time,
            duration_cast_operator=duration_cast_operator,
        )
        entities = self.entity_consensus(entities, len(transcripts))
        return self.apply_filters(entities)

    def utility(self, input: Input, output: Output) -> List[BaseEntity]:
        """
        Produces Duckling entities, runs with a :ref:`Workflow's run<workflow_run>` method.

        :param argbrightons: Expects a tuple of :code:`Tuple[natural language for parsing entities, reference time in milliseconds, locale]`
        :type args: Tuple(Union[str, List[str]], int, str)
        :return: A list of duckling entities.
        :rtype: List[BaseEntity]
        """
        transcripts = input.transcripts
        self.reference_time = input.reference_time
        self.locale = input.locale or self.locale
        use_latent = input.latent_entities

        return self.parse(
            transcripts,
            locale=self.locale,
            reference_time=self.reference_time,
            use_latent=use_latent,
            intents=output.intents,
        )

    def transform(self, training_data: pd.DataFrame) -> pd.DataFrame:
        """
        Transform training data.

        :param training_data: Training data.
        :type training_data: pd.DataFrame
        :return: Transformed training data.
        :rtype: pd.DataFrame
        """
        if not self.use_transform:
            return training_data

        logger.debug(f"Transforming dataset via {self.__class__.__name__}")
        logger.disable("dialogy")
        training_data = training_data.copy()
        if self.output_column not in training_data.columns:
            training_data[self.output_column] = None

        for i, row in tqdm(training_data.iterrows(), total=len(training_data)):
            reference_time = row[self.reference_time_column]
            if isinstance(reference_time, str):
                reference_time = int(
                    datetime.fromisoformat(reference_time).timestamp() * 1000
                )
            elif pd.isna(reference_time):
                continue
            elif not isinstance(reference_time, int):
                raise TypeError(
                    f"{reference_time=} should be isoformat date or unix timestamp integer."
                )
            transcripts = self.make_transform_values(row[self.input_column])
            entities = self.parse(
                transcripts,
                lang_detect_from_text(self.input_column),
                reference_time=reference_time,
                use_latent=self.activate_latent_entities,
            )
            if row[self.output_column] is None or pd.isnull(row[self.output_column]):
                training_data.at[i, self.output_column] = entities
            else:
                training_data.at[i, self.output_column] = (
                    row[self.output_column] + entities
                )
        logger.enable("dialogy")
        return training_data



"""
.. _PeopleEntity:

People Entity
==============

Provides the entity class for representing number of people in natural language. This entity is obtained via :ref:`DucklingPlugin<DucklingPlugin>`.

Plugin Walkthrough
------------------

.. ipython::

    In [1]: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["people"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: duckling_plugin.parse("two kids")

Workflow Integration
--------------------

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.workflow import Workflow

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["people"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: _, output = workflow.run(Input(utterances="two kids"))

    In [5]: output
"""
from __future__ import annotations

from typing import Any, Dict, Optional

import attr

from dialogy import constants as const
from dialogy.types.entity.base_entity import BaseEntity


@attr.s
class PeopleEntity(BaseEntity):
    """
    A variant of numerical entity which addresses collections of people.

    Example sentences that contain the entity are:
    - "N people", where N is a whole number.
    - "A couple" ~ 2 people.
    - "I have triplets" ~ 3 children.
    """

    unit = attr.ib(type=str, default="", validator=attr.validators.instance_of(str))
    entity_type: Optional[str] = attr.ib(default="people", order=False)

    @classmethod
    def from_duckling(cls, d: Dict[str, Any], alternative_index: int) -> PeopleEntity:
        value = d[const.VALUE][const.VALUE]
        return cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            latent=d[const.LATENT],
            values=[{const.VALUE: value}],
            unit=d[const.VALUE][const.UNIT],
        )



"""
Trains a calibraation model. This contains two models:
- Vectorizer: TfIdf
- Classifier: XGBoostRegressor
"""

import json
import math
import pickle
import traceback
from typing import Any, Dict, List, Optional, Tuple

import jiwer
import numpy as np
import pandas as pd
from loguru import logger
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from tqdm import tqdm
from xgboost import XGBRegressor

from dialogy import constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.types import Transcript, Utterance, utterances
from dialogy.utils import normalize


class FeatureExtractor(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.vectorizer = TfidfVectorizer()

    def fit(self, df: pd.DataFrame, y: Any = None) -> Any:
        texts = []
        for _, row in tqdm(df.iterrows()):
            real_transcript = json.loads(row["tag"])["text"]
            texts.append(real_transcript)
            alts = json.loads(row["data"])
            if alts not in [[], [None]]:
                for alt in alts[0]:
                    texts.append(alt["transcript"])

        logger.debug("Step 1/2: Training vectorizer model")
        self.vectorizer.fit(texts)
        return self

    def features(self, alternatives: List[Dict[str, Any]]) -> List[List[float]]:
        features = []
        for alt in alternatives:
            try:
                l = len(alt["transcript"].split())
                features.append(
                    self.vectorizer.transform([alt["transcript"]]).todense().tolist()[0]
                    + [
                        alt["am_score"] / l,
                        alt["lm_score"] / l,
                        alt["transcript"].count("UNK"),
                        l,
                        alt["am_score"] / math.log(1 + l),
                        alt["lm_score"] / math.log(1 + l),
                        alt["am_score"] / math.sqrt(l),
                        alt["am_score"] / math.sqrt(l),
                    ]
                )
            except Exception as error:
                logger.error(f"{error}\n{traceback.format_exc()}")

        return features

    def transform(self, df: pd.DataFrame) -> Tuple[Any, Any]:
        features, targets = [], []
        for _, row in tqdm(df.iterrows()):
            real_transcript = json.loads(row["tag"])["text"]
            utterances: List[Utterance] = json.loads(row["data"])
            if utterances not in [[], [None]]:
                for utterance in utterances:
                    features.append(self.features(utterance))
                    targets += [
                        jiwer.wer(real_transcript, alternative["transcript"])
                        for alternative in utterance
                    ]
        return np.squeeze(np.array(features)), targets


class CalibrationModel(Plugin):
    """
    .. _calibration_plugin:

    This plugin provides a calibration model that sits between ASR and SLU.
    It trains a model that learn to classify alternatives from the text and
    AM, LM score. Bad alternatives are removed before training SLU and during
    inference.
    """

    def __init__(
        self,
        threshold: float,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        debug: bool = False,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = const.ALTERNATIVES,
        use_transform: bool = False,
        model_name: str = "calibration.pkl",
    ) -> None:
        super().__init__(
            dest=dest,
            guards=guards,
            debug=debug,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
        )
        self.extraction_pipeline = FeatureExtractor()
        self.clf = XGBRegressor(n_jobs=1)
        self.threshold = threshold
        self.model_name = model_name

    def train(self, df: pd.DataFrame) -> None:
        """
        Trains the calibration pipeline.

        :param df: dataframe to train on. Should be a valid transcrition tagging job.
        :param model_name: Saves the pipline as {model_name}.pkl
        :type df: pd.DataFrame
        :type model_name: str
        """
        X, y = self.extraction_pipeline.fit_transform(df)
        logger.debug("Step 2/2: Training regressor model")
        self.clf.fit(X, y)
        self.save(self.model_name)

    def predict(self, alternatives: Utterance) -> Any:
        return self.clf.predict(
            np.array(self.extraction_pipeline.features(alternatives))
        )

    def filter_asr_output(self, utterances: List[Utterance]) -> List[Utterance]:
        """
        Filters outputs from ASR based on calibration model prediction.

        :param asr_output: output dictionary from ASR. Should have an _alternatives_
                            key.
        :type utterances: List[Utterance]
        :return: Filtered alternatives, in the same format as input.
        :rtype: Dict[str, Any]
        """
        filtered_utterances = []
        for utterance in utterances:
            filtered_alternatives = []
            prediction = self.predict(utterance)
            for alternative, wer in zip(utterance, prediction):
                if wer < self.threshold:
                    filtered_alternatives.append(alternative)
            filtered_utterances.append(filtered_alternatives)
        return filtered_utterances

    def transform(self, training_data: pd.DataFrame) -> pd.DataFrame:
        # filters df alternatives and feeds into merge_asr_output,
        # doesn't change training_data schema
        training_data["use"] = True
        logger.debug("Transforming training data.")
        for i, row in training_data.iterrows():
            asr_output = None
            try:
                asr_output = json.loads(row[self.input_column])
                if asr_output:
                    filtered_asr_output = self.filter_asr_output(asr_output)
                    training_data.iloc[i][self.input_column] = filtered_asr_output
                else:
                    training_data.loc[i, "use"] = False
            except Exception as error:  # pylint: disable=broad-except
                training_data.loc[i, "use"] = False
                logger.error(f"{error} -- {asr_output}\n{traceback.format_exc()}")

        training_data_ = training_data[training_data.use].copy()
        training_data_.drop("use", axis=1, inplace=True)
        discarded_data = len(training_data) - len(training_data_)
        if discarded_data:
            logger.debug(
                f"Discarding {discarded_data} samples because the alternatives couldn't be parsed."
            )
        return training_data_

    def inference(
        self, transcripts: List[str], utterances: List[Utterance]
    ) -> List[str]:
        transcript_lengths: List[int] = [
            len(transcript.split()) for transcript in transcripts
        ]
        average_word_count: float = (
            sum(transcript_lengths) / len(transcript_lengths) if transcripts else 0.0
        )

        # We want to run this plugin if transcripts have more than WORD_THRESHOLD words
        # below that count, WER is mostly high. We expect this plugin to override
        # a classifier's prediction to a fallback label.
        # If the transcripts have less than WORD_THRESHOLD words, we will always predict the fallback label.
        if average_word_count <= const.WORD_THRESHOLD:
            return transcripts

        return normalize(self.filter_asr_output(utterances))

    def save(self, fname: str) -> None:
        pickle.dump(self, open(fname, "wb"))

    def utility(self, input: Input, _: Output) -> Any:
        return self.inference(
            input.transcripts, input.utterances
        )  # pylint: disable=no-value-for-parameter

    def validate(self, df: pd.DataFrame) -> bool:
        """
        Return if `df` is a valid trascription tagging job
        should return `False` for intent tagging jobs.
        example : `'{"text": "I want to change and set my <INAUDIBLE>", "type": "TRANSCRIPT"}'`

        Sharp bits:
        - All rows in df should have same format. We just consider
         the first row for sanity checks.

        :param df: Input dataframe.
        :type df: pd.DataFrame

        :return: (bool) if the dataframe is valid for training calibration model.
        :rtype: bool
        """
        required_keys = ["text", "type"]
        tagged_data = json.loads(df.iloc[0]["tag"])
        if all(key in tagged_data for key in required_keys):
            if tagged_data["type"] == "TRANSCRIPT":
                return True
        return False



"""
There are few types of significance for an SLU project

1. :ref:`Intent<intent>`
2. :ref:`Entity<base_entity>`
3. :ref:`Slot<slot>`
4. :ref:`Signal<signal>`

.. warning:: This section is a work in progress.

An :ref:`Intent<intent>` tells us the general meaning of a sentence.
A set :ref:`Entity<base_entity>` are significant tokens that can be acted upon by either
participants within a conversation.

We use :ref:`slots<slot>` within :ref:`intents<intent>` to associate relationship between :ref:`intents<intent>` and
:ref:`entities<base_entity>`. You can consider examples like: Reservation of a air ticket requires:

- destination
- origin
- travel date
- travel time
- travel type (round trip, one-way)
- number of passengers.

So these :ref:`entities<base_entity>` would be filled within their respective :ref:`slots<slot>` already expected by the
reservation :ref:`intents<intent>`.
"""
from dialogy.types.entity.amount_of_money import CurrencyEntity
from dialogy.types.entity.base_entity import BaseEntity, entity_synthesis
from dialogy.types.entity.credit_card_number import CreditCardNumberEntity
from dialogy.types.entity.deserialize import deserialize_duckling_entity
from dialogy.types.entity.duration import DurationEntity
from dialogy.types.entity.keyword import KeywordEntity
from dialogy.types.entity.numerical import NumericalEntity
from dialogy.types.entity.people import PeopleEntity
from dialogy.types.entity.time import TimeEntity
from dialogy.types.entity.time_interval import TimeIntervalEntity
from dialogy.types.intent import Intent
from dialogy.types.plugin import PluginFn
from dialogy.types.signal.signal import Signal
from dialogy.types.slots import Slot
from dialogy.types.utterances import Alternative, Transcript, Utterance



"""
.. _keyword_entity:

Module provides access to an entity type (class) to contain keyword extraction based entites.
These entities originate from token lookups, regex, etc.

Import classes:
    - KeywordEntity
"""
from __future__ import annotations

from typing import Any, Dict, Optional

import attr

from dialogy import constants as const
from dialogy.types.entity.base_entity import BaseEntity


@attr.s
class KeywordEntity(BaseEntity):
    """
    Use this type for handling keyword based extractions where presence of specific tokens in the ASR
    is enough for detection.
    """

    _meta = attr.ib(type=Dict[str, str], default=attr.Factory(dict))



"""
This module was created in response to: https://github.com/Vernacular-ai/dialogy/issues/9
we will ship functions to assist normalization of ASR output, we will refer to these as Utterances.
"""
import itertools
from functools import partial
from typing import Any, Callable, Dict, List

from dialogy import constants as const
from dialogy.types.utterances import Transcript, Utterance


def dict_get(prop: str, obj: Dict[str, Any]) -> Any:
    """
    Get value of prop within obj.

    This simple function exists to facilitate a partial function defined :ref:`here<is_utterance>`.

    :param prop: A property within a :code:`dict`.
    :type prop: str
    :param obj: A :code:`dict`.
    :type obj: Dict[str, Any]
    :return: Value of a property within a :code:`dict`.
    :rtype: Any
    """
    return obj[prop]


def is_list(input_: Any) -> bool:
    """
    Check type of :code:`input`

    :param input_: Any arbitrary input
    :type input_: Any
    :return: True if :code:`input` is a :code:`list` else False
    :rtype: True
    """
    return isinstance(input_, list)


def is_each_element(
    type_: type, input_: List[Any], transform: Callable[[Any], Any] = lambda x: x
) -> bool:
    """
    Check if each element in a list is of a given type.

    :param type_: Expected :code:`Type` of each element in the :code:`input_` which is a :code:`list`.
    :type type_: Type
    :param input_: A :code:`list`.
    :type input_: List[Any]
    :param transform: We may apply some transforms to
        each element before making these checks. This is to check if a certain key
        in a Dict matches the expected type. In case this is not needed,
        leave the argument unset and an identity transform is applied. Defaults to lambda x:x.
    :type transform: Callable[[Any], Any]
    :return: Checks each element in a list to match :code:`type_`, if any element fails the check,
        this returns False, else True.
    :rtype: bool
    """
    return all(isinstance(transform(item), type_) for item in input_)


def is_utterance(maybe_utterance: Any, key: str = const.TRANSCRIPT) -> bool:
    """
    .. _is_utterance:

    Check input to be of `List[List[Dict]]`.


    .. code-block: json

        [[{"transcript": "hi"}]]


    .. ipython:: python

        from dialogy.utils.normalize_utterance import is_utterance

        # 1. :code:`List[List[Dict[str, str]]]`
        is_utterance([[{"transcript": "this"}, {"transcript": "works"}]])

        # 2. key is configurable
        is_utterance([[{"text": "this"}, {"text": "works"}]], key="text")

        # 3. Hope for everything else... you have a mastercard.
        # Or use this lib, works just fine 🍷.
        is_utterance([{"transcript": "this"}, {"transcript": "doesn't"}, {"transcript": "work"}])

    :param maybe_utterance: Arbitrary input.
    :type maybe_utterance: Any
    :param key: The key within which transcription string resides.
            Defaults to :code:`const.TRANSCRIPT`.
    :type key: str
    :return: True if the inputs is :code:`List[List[Dict[str, str]]]`, else False.
    :rtype: bool
    """
    dict_get_key = partial(dict_get, key)
    try:
        return all(
            is_each_element(str, alternatives, transform=dict_get_key)
            for alternatives in maybe_utterance
        )
    except KeyError:
        return False
    except TypeError:
        return False


def is_unsqueezed_utterance(maybe_utterance: Any, key: str = const.TRANSCRIPT) -> bool:
    """
    Check input to be of `List[Dict]`.

    .. ipython:: python

        from dialogy.utils.normalize_utterance import is_unsqueezed_utterance

        # 1. This fails
        is_unsqueezed_utterance([[{"transcript": "this"}, {"transcript": "works"}]])

        # 2. key is configurable
        is_unsqueezed_utterance([{"text": "this"}, {"text": "works"}], key="text")

    :param maybe_utterance: Arbitrary type input.
    :type maybe_utterance: Any
    :param key: The key within which transcription string resides.
    :type key: str, Defaults to const.TRANSCRIPT.
    :return: True, if the input is of type :code:`List[Dict[str, Any]]` else False.
    :rtype: bool
    """
    dict_get_key = partial(dict_get, key)
    try:
        return is_each_element(str, maybe_utterance, transform=dict_get_key)
    except KeyError:
        return False
    except TypeError:
        return False


def is_list_of_string(maybe_utterance: Any) -> bool:
    """
    Check input to be of `List[str]`.

    .. ipython:: python

        from dialogy.utils.normalize_utterance import is_list_of_string
        is_list_of_string(["this", "works"])

    :param maybe_utterance: Arbitrary input.
    :type maybe_utterance: Any
    :return: True if :code:`maybe_utterance` is a :code:`str`.
    :rtype: bool
    """
    try:
        return is_each_element(str, maybe_utterance)
    except TypeError:
        return False


def is_string(maybe_utterance: Any) -> bool:
    """
    Check input's type is `str`.

    :param maybe_utterance: Arbitrary type input.
    :type maybe_utterance: Any
    :return: True if :code:`maybe_utterance` is a :code:`str`, else False.
    :rtype: bool
    """
    return isinstance(maybe_utterance, str)


def utterance2alternatives(
    utterances: List[Utterance], key: str = const.TRANSCRIPT
) -> List[Transcript]:
    """
    Convert a list of utterances to a list of alternatives.
    """
    return [
        " ".join([alternative[key] for alternative in alternatives])
        for alternatives in itertools.product(*utterances)
    ]


def normalize(maybe_utterance: Any, key: str = const.TRANSCRIPT) -> List[str]:
    """
    Adapt various non-standard ASR alternative forms.

    The output will be a list of strings since models will expect that.

    .. _normalize:

    .. ipython:: python

        In [1]: from dialogy.utils.normalize_utterance import normalize
        In [2]: # A popular case
        In [3]: normalize([[{"transcript": "this"}, {"transcript": "works"}]])

    .. ipython:: python

        In [3]: # A case with multiple utterances
        In [4]: normalize([
                    [{"transcript": "hello hello?", "transcript": "yellow yellow?"}],
                    [{"transcript": "I wanted to check"}],
                    [{"transcript": "if you have space for us?"}]
                ])

    .. ipython:: python

        In [5]: normalize([{"transcript": "I wanted to know umm hello?"}])

    .. ipython:: python

        In [6]: normalize(["I wanted to know umm hello?"])

    .. ipython:: python

        In [7]: normalize("I wanted to know umm hello?")

    :param maybe_utterance: Arbitrary input.
    :type maybe_utterance: Any
    :param key: A string to be looked into :code:`List[List[Dict[str, str]]]`, :code:`List[Dict[str, str]]` type inputs.
    :type key: str
    :return: A flattened list of strings parsed from various formats.
    :rtype: List[str]
    :raises:
        TypeError: If :code:`maybe_utterance` is none of the expected types.
    """
    if maybe_utterance is None or not maybe_utterance:
        return []

    if is_utterance(maybe_utterance):
        return utterance2alternatives(maybe_utterance)

    if is_unsqueezed_utterance(maybe_utterance):
        return [alternative[key] for alternative in maybe_utterance]

    if is_list(maybe_utterance) and is_list_of_string(maybe_utterance):
        return [utterance for utterance in maybe_utterance]

    if is_string(maybe_utterance):
        return [maybe_utterance]

    else:
        raise TypeError(
            f"The input {maybe_utterance} does not belong to any of the expected types:"
            " List[List[Dict[str, Any]]], List[Dict[str, Any]], List[str] or str."
        )



"""
.. _TimeEntity:

Time Entity
============

Provides the entity class for representing time in natural language. This entity is obtained via :ref:`DucklingPlugin<DucklingPlugin>`.

Plugin Walkthrough
------------------

.. ipython::

    In [1]: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: duckling_plugin.parse("tomorrow")

    In [4]: duckling_plugin.parse("monday")

    In [7]: duckling_plugin.parse("27th jan")

Workflow Integration
--------------------

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.workflow import Workflow

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: _, output = workflow.run(Input(utterances=["tomorrow", "monday", "27th jan"]))

    In [5]: output
"""
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional

import attr
import pydash as py_

from dialogy import constants as const
from dialogy.types.entity.base_entity import BaseEntity


@attr.s
class TimeEntity(BaseEntity):
    """
    Entities that can be parsed to obtain date, time or datetime values.

    Example sentences that contain the entity are:
    - "I have a flight at 6 am."
    - "I have a flight at 6th December."
    - "I have a flight at 6 am today."

    Attributes:
    - `grain` tells us the smallest unit of time in the utterance
    """

    origin: str = attr.ib(default="value")
    dim: str = attr.ib(default="time")
    grain: str = attr.ib(default=None, validator=attr.validators.instance_of(str))

    def get_value(self) -> Any:
        """
        Return the date string in ISO format from the dictionary passed

        .. code-block:: python

            date = {
                "value": "2021-04-17T16:00:00.000+05:30",
                "grain": "hour",
                "type": "value"
            }

        :param date: Dictionary which stores the datetime in ISO format, grain and type
        :type date: Dict[str, str]
        :return: :code:`date["value"]`
        :rtype: Optional[datetime]
        """
        entity_date_value = super().get_value()
        return datetime.fromisoformat(entity_date_value)

    def collect_datetime_values(self) -> List[datetime]:
        """
        Collect all datetime values from the entity

        :return: List of datetime values
        :rtype: List[str]
        """
        return [datetime.fromisoformat(value[const.VALUE]) for value in self.values]

    def is_uniq_date_from_values(self) -> bool:
        """
        Check a list has a unique day

        Where day is the date number for a month.

        :return: True if there is a unique day in all datetime values
        :rtype: bool
        """
        dates = self.collect_datetime_values()
        return len(py_.uniq(dates)) == 1

    def is_uniq_day_from_values(self) -> bool:
        """
        Check a list has a unique weekday

        Where weekday ranges from 0 for Monday till 6 for Sunday
        Ex: for "2021-04-17T16:00:00.000+05:30", the weekday is 5 (Saturday)

        Duckling may return 3 datetime values in chronological order. For cases where the utterance is "Monday",
        (or any weekday for that matter) we will get 3 values, each value a week apart.
        We pair these dates and check the difference is 1 week long (7 days).

        For example:

        .. code-block:: python

            # Say, our 3 date values are:
            date values = [21-04-2021, 28-04-2021, 05-05-2021]

            # then we generate pairs as:
            [(21-04-2021, 28-04-2021), (28-04-2021, 05-05-2021)]

        If the difference between these dates is divisible by seven, then the input was a weekday.

        :return: True if there is a unique weekday in all datetime values
        :rtype: bool
        """
        dates = self.collect_datetime_values()
        return all(
            abs((next - prev).days) % 7 == 0
            for (prev, next) in zip(dates[:-1], dates[1:])
        )

    def set_entity_type(self) -> None:
        """
        Returns the type for this entity

        To pinpoint a time for any action, we need both DATE (either explicit or inferred) and TIME.
        We can get these in 3 possible manners:

        - both DATE and TIME, the entity type here will be DATETIME
            - Positive examples: tonight, tomorrow 4pm, 7th april 2019 6pm, last 3 hours (Because it tells us about the date as well, that is today)
            - Negative examples: Yesterday I went to park at 3pm (because the user did not say yesterday and 3pm together)
        - only DATE, the entity type here will be DATE
            - Positive examples: last month, last year, last 3 months, last 6 months, tomorrow, 27th oct 1996, 1st October 2019 till yesterday, yesterday, today, tomorrow, October, 2019, 27th October 1996 to 29th October 1996 etc
            - Negative examples: month, year, 6 months, 3 months (Since these do not have a fixed start and end date, it could have been 6 months from today, 6 months from tomorrow, previous 6 months. Therefore it only tells us a duration.)
        - only TIME, the entity type here will be TIME
            - Positive examples: 3pm, 5pm, night, morning, etc
            - Negative examples: tonight (because the date part today is in it), tomorrow night, 6th April 5pm, 7 hours (duration because they didn't mention a start/end time. It could be 7 hours from any time) etc

        if we get the grain as one of ("day", "week", "month", "quarter", "year"),
        then the entity type is DATE
        if we get grain as one of ("hour", "second", "minute") and we only have 1 value in `self.values`,
        then the entity type is DATETIME.
        if we get grain as one of ("hour", "second", "minute") and we have either a unique date or a unique weekday,
        then the entity type is DATETIME.
        else the entity type is TIME.

        :return: one of (date, time, datetime)
        :rtype: str
        """
        if self.grain in const.DATE_UNITS:
            self.entity_type = const.DATE
        elif self.grain in const.TIME_UNITS and len(self.values) == 1:
            self.entity_type = const.DATETIME
        elif self.is_uniq_date_from_values() or self.is_uniq_day_from_values():
            self.entity_type = const.DATETIME
        elif len(self.values) > 0:
            self.entity_type = const.TIME

    @classmethod
    def from_duckling(cls, d: Dict[str, Any], alternative_index: int) -> TimeEntity:
        time_entity = cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            latent=d[const.LATENT],
            values=d[const.VALUE][const.VALUES],
            grain=d[const.VALUE][const.VALUES][0][const.GRAIN],
        )
        time_entity.set_entity_type()
        return time_entity



"""
.. _Intent:

Intent Classification
======================

Automatic categorization of texts into known classes. When solving SLU problems
the objective is to answer *What was meant?* instead of *"What was said?"*. 

For an example utterance *"I lost my wallet"* a bank's support executives would 
understand that they must follow the operating procedure for blocking all transaction 
accounts. We want to train machines to have the ability to interpret these cues.

We call the classes of *meaning* as :code:`Intent`.

We can obtain intents from :ref:`XLMRMultiClass<XLMRMultiClass>` and :ref:`MLPMultiClass<MLPMultiClass>` plugin.
"""
from __future__ import annotations

from typing import Any, Dict, List, Optional

import attr

from dialogy.types import BaseEntity
from dialogy.types.slots import Rule, Slot
from dialogy.utils.logger import logger


@attr.s
class Intent:
    """
    An instance of this class contains the name of the action associated with a body of text.
    """

    # The name of the intent to be used.
    name: str = attr.ib(kw_only=True, order=False)
    """
    The description of an intent separated by '_'
    """

    # The confidence of this intent being present in the utterance.
    score: float = attr.ib(kw_only=True, default=0.0, order=True)
    """
    A positive real number that represents confidence of this intent
    being the meaning of an utterance. Higher is better.
    
    Should be a value between 0 and 1.
    """

    # In case of an ASR, `alternative_index` points at one of the nth
    # alternatives that help in predictions.
    alternative_index: Optional[int] = attr.ib(kw_only=True, default=None, order=False)
    """
    Tells us the index of transcript that yeilds this intent. Since our featurizer
    concatenates all transcripts, we currently don't have a use for it.

    If we were to predict the intent for each transcript separately and vote, in those
    cases it would be helpful to log this property.
    """

    # Trail of functions that modify the attributes of an instance.
    parsers: List[str] = attr.ib(kw_only=True, factory=list, order=False, repr=False)
    """
    Contains a list of plugins that have created, updated this Intent.
    """

    # Container for holding `List[BaseEntity]`.
    slots: Dict[str, Slot] = attr.ib(
        kw_only=True, factory=dict, order=False, repr=False
    )
    """
    Placeholders for :ref:`entities <BaseEntity>` that are relevant to this intent. 
    More details are available in the doc for :ref:`slots <Slot>`.
    """

    def apply(self, rules: Rule) -> Intent:
        """
        Create slots using rules.

        .. _ApplySlot:

        An intent can hold different entities within associated slot-types.
        We parse rules that describe this mapping and create placeholders within an intent for
        holding entities found.

        An example rule would look like:

        .. code:: python

            rules = {
                "wake_up_alarm": {          # name of the intent.
                    "datetime_slot": [      # "datetime_slot" is the only slot associated with this intent.
                        "date",             # Only "date", "time", "datetime" entities can fill the "datetime_slot"
                        "time",
                        "datetime"
                    ]
                },
                "ticket_booking": {         # name of the intent.
                    "origin": [             # A slot that will fill only location type entities.
                        "location"          # The entity types that can be filled in the slot.
                    ],
                    "destination": [        # Another slot that will fill only location type entities.
                        "location"
                    ],
                    "start_date": [
                        "date"
                    ],
                    "end_date": [
                        "date"
                    ],
                    "passengers": [
                        "people",
                        "number"
                    ],
                    "budget": [
                        "amount-of-money"
                    ]
                }
            }

        :param rules: A configuration for slot names and entity types associated with this instance of Intent.
        :type rules: Rule
        :return: The calling Intent is modified to have :ref:`Slots <slot>`.
        :rtype: Intent
        """
        rule = rules.get(self.name)
        if not rule:
            return self

        for slot_name, entity_types in rule.items():
            if isinstance(entity_types, str):
                entity_type = entity_types
                self.slots[slot_name] = Slot(
                    name=slot_name, types=[entity_type], values=[]
                )
            elif isinstance(entity_types, list) and all(
                isinstance(type_, str) for type_ in entity_types
            ):
                self.slots[slot_name] = Slot(
                    name=slot_name, types=entity_types, values=[]
                )
            else:
                raise TypeError(
                    f"Expected entity_types={entity_types} in the rule"
                    f" {rule} to be a List[str] but {type(entity_types)} was found."
                )

        return self

    def add_parser(self, plugin: Any) -> Intent:
        """
        Update parsers with the plugin name

        This is to identify the progression in which the plugins were applied to an intent.
        This only helps in debugging and has no production utility.

        :param plugin: The class that modifies this instance. Preferably should be a plugin.
        :type plugin: Plugin
        :return: Calling instance with modifications to :code:`parsers` attribute.
        :rtype: Intent
        """
        plugin_name = plugin if isinstance(plugin, str) else plugin.__class__.__name__
        self.parsers.append(plugin_name)
        return self

    def fill_slot(self, entity: BaseEntity, fill_multiple: bool = False) -> Intent:
        """
        Update :code:`slots[slot_type].values` with a single entity.

        .. _FillSlot:

        We will explore the possibility of sharing/understanding the meaning of multiple entities
        in the same slot type.

        There maybe cases where we want to fill multiple entities of the same type within a slot.
        In these cases :code:`fill_multiple` should be set to True.

        The :ref:`RuleBasedSlotFillerPlugin <rule_slot_filler>` has a detailed demo for this.

        :param entity: The entity to be checked for support in the calling intent's slots.
        :type entity: BaseEntity
        :param fill_multiple:
        :type fill_multiple: bool
        :return: The calling Intent with modifications to its slots.
        :rtype: Intent
        """
        logger.debug("Looping through slot_names for each entity.")
        logger.debug(f"intent slots: {self.slots}")
        for slot_name, slot in self.slots.items():
            logger.debug(f"slot_name: {slot_name}")
            logger.debug(
                f"slot type: {slot.types}",
            )
            logger.debug(
                f"entity type: {entity.entity_type}",
            )
            if entity.entity_type in slot.types:
                if fill_multiple:
                    logger.debug(f"filling {entity} into {self.name}.")
                    self.slots[slot_name].add(entity)
                    return self

                if not self.slots[slot_name].values:
                    logger.debug(f"filling {entity} into {self.name}.")
                    self.slots[slot_name].add(entity)
                else:
                    logger.debug(
                        f"removing {entity} from {self.name}, because the slot was filled previously. "
                        "Use fill_multiple=True if this is not required.",
                        entity,
                        self.name,
                    )
                    self.slots[slot_name].clear()
        return self

    def cleanup(self) -> None:
        """
        Remove slots that were not filled.

        .. _CleanupSlot:
        """
        slot_names = list(self.slots.keys())
        for slot_name in slot_names:
            if not self.slots[slot_name].values:
                del self.slots[slot_name]

    def json(self) -> Dict[str, Any]:
        """
        Convert the object to a dictionary.

        .. ipython:: python

            from dialogy.types.intent import Intent

            intent = Intent(name="special", score=0.8)
            intent.json()
        """
        return {
            "name": self.name,
            "alternative_index": self.alternative_index,
            "score": self.score,
            "parsers": self.parsers,
            "slots": [slot.json() for slot in self.slots.values()],
        }



"""
.. _mlp_classifier:

This module provides a trainable MLP classifier.
"""
import ast
import operator
import os
from typing import Any, Dict, List, Optional

import joblib
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import f1_score, make_scorer
from sklearn.model_selection import GridSearchCV
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from tqdm import tqdm

tqdm.pandas()

import dialogy.constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.plugins.text.classification.tokenizers import identity_tokenizer
from dialogy.types import Intent
from dialogy.utils import load_file, logger, save_file


class MLPMultiClass(Plugin):
    """
    This plugin provides a classifier based on sklearn's MLPClassifier.

    .. _MLPMultiClass:
    """

    def __init__(
        self,
        model_dir: str,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        debug: bool = False,
        threshold: float = 0.1,
        score_round_off: int = 5,
        purpose: str = const.TRAIN,
        fallback_label: str = const.ERROR_LABEL,
        data_column: str = const.DATA,
        label_column: str = const.LABELS,
        args_map: Optional[Dict[str, Any]] = None,
        skip_labels: Optional[List[str]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:

        super().__init__(dest=dest, guards=guards, debug=debug)
        self.model_pipeline: Any = None
        self.fallback_label = fallback_label
        self.data_column = data_column
        self.label_column = label_column
        self.mlp_model_path = os.path.join(model_dir, const.MLPMODEL_FILE)
        self.threshold = threshold
        self.skip_labels = set(skip_labels or set())
        self.purpose = purpose
        self.round = score_round_off
        if args_map and (
            const.TRAIN not in args_map
            or const.TEST not in args_map
            or const.PRODUCTION not in args_map
        ):
            raise ValueError(
                f"Attempting to set invalid {args_map=}. "
                "It is missing some of {const.TRAIN}, {const.TEST}, {const.PRODUCTION} in configs."
            )
        self.args_map = args_map
        self.kwargs = kwargs or {}
        try:
            if os.path.exists(self.mlp_model_path):
                self.init_model()
        except EOFError:
            logger.error(
                f"Plugin {self.__class__.__name__} Failed to load MLPClassifier Model from {self.mlp_model_path}. "
                "Ignore this message if you are training but if you are using this in production or testing, then this is serious!"
            )

    def init_model(self) -> Dict[str, Any]:
        """
        Initialize the model if artifacts are available.
        """
        if os.path.exists(self.mlp_model_path):
            self.load()
            return {}
        args = (
            self.args_map[self.purpose]
            if self.args_map and self.purpose in self.args_map
            else {}
        )

        self.model_pipeline = Pipeline(
            [
                (
                    const.TFIDF,
                    TfidfVectorizer(
                        tokenizer=identity_tokenizer,
                        stop_words=const.STOPWORDS,
                        lowercase=const.TFIDF_LOWERCASE,
                        ngram_range=const.DEFAULT_NGRAM,
                    ),
                ),
                (
                    const.MLP,
                    MLPClassifier(
                        random_state=const.MLP_RANDOMSTATE,
                        max_iter=(
                            args[const.NUM_TRAIN_EPOCHS]
                            if args
                            else const.MLP_DEFAULT_TRAIN_EPOCHS
                        ),
                        verbose=True,
                    ),
                ),
            ]
        )
        USE = "use"
        if args and args[const.USE_GRIDSEARCH][USE]:
            GRID = self.get_gridsearch_grid(**args[const.USE_GRIDSEARCH][const.PARAMS])
            self.model_pipeline = GridSearchCV(
                estimator=self.model_pipeline,
                param_grid=GRID,
                cv=args[const.USE_GRIDSEARCH][const.CV],
                n_jobs=const.GRIDSEARCH_WORKERS,
                verbose=args[const.USE_GRIDSEARCH][const.VERBOSE_LEVEL],
                scoring=make_scorer(f1_score, average=const.GRID_SCORETYPE),
                return_train_score=True,
            )

        return args

    def get_gridsearch_grid(self, **kwargs: Any) -> List[Dict[str, List[Any]]]:
        """
        Gets gridsearch hyperparameters for the model in proper grid params format.

        Raises:
            ValueError: If a gridsearch parameter doesn't exist in sklearns TFIDF and MLPClassifier modules.
        """
        grid_params: Dict[str, List[Any]] = {}
        for k, v in kwargs.items():
            if (
                k not in self.model_pipeline[const.TFIDF].get_params().keys()
                and k not in self.model_pipeline[const.MLP].get_params().keys()
            ):
                raise ValueError(
                    f"Hyperparam defined for gridsearch {k} doesn't exist,\nRefer: https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html\nand https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html"
                )
            if k != const.NGRAM_RANGE:
                grid_params[f"{const.MLP}__{k}"] = self.get_formatted_gridparams(v)
            else:
                grid_params[f"{const.TFIDF}__{k}"] = self.get_formatted_gridparams(v)

        return [grid_params]

    @staticmethod
    def get_formatted_gridparams(params: List[Any]) -> List[Any]:
        """
        Gets the valid parameters for the gridsearch.

        Args:
            values: The values to be validated.

        Returns:
            The valid parameters.
        """
        valid_params: List[Any] = []
        for p in params:
            try:
                valid_params.append(ast.literal_eval(p))
            except ValueError:
                valid_params.append(p)
        return valid_params

    @property
    def valid_mlpmodel(self) -> bool:
        return hasattr(self.model_pipeline, "classes_")

    def inference(self, texts: Optional[List[str]]) -> List[Intent]:
        """
        Predict the intent of a list of texts.

        :param texts: A list of strings, derived from ASR transcripts.
        :type texts: List[str]
        :raises AttributeError: In case the model isn't of sklearn pipeline or gridsearchcv.
        :return: A list of intents corresponding to texts.
        :rtype: List[Intent]
        """
        logger.debug(f"Classifier input:\n{texts}")
        fallback_output = Intent(name=self.fallback_label, score=1.0).add_parser(self)

        if self.model_pipeline is None:
            logger.error(f"No model found for plugin {self.__class__.__name__}!")
            return [fallback_output]

        if not texts:
            return [fallback_output]

        if not isinstance(self.model_pipeline, Pipeline) and not isinstance(
            self.model_pipeline, GridSearchCV
        ):
            raise AttributeError(
                "Seems like you forgot to "
                f"save the {self.__class__.__name__} plugin."
            )

        try:
            probs_and_classes = sorted(
                zip(
                    self.model_pipeline.predict_proba(texts)[0],
                    self.model_pipeline.classes_,
                ),
                key=operator.itemgetter(0),
                reverse=True,
            )
        except NotFittedError:
            logger.error(f"{self.__class__.__name__} model not trained yet!")
            return [fallback_output]

        return [
            Intent(name=intent, score=round(score, self.round)).add_parser(self)
            for score, intent in probs_and_classes
        ]

    def validate(self, training_data: pd.DataFrame) -> bool:
        """
        Validate the training data is in the appropriate format

        :param training_data: A pandas dataframe containing at least list of strings and corresponding labels.
        :type training_data: pd.DataFrame
        :return: True if the dataframe is valid, False otherwise.
        :rtype: bool
        """
        if training_data.empty:
            logger.error("Training dataframe is empty.")
            return False

        for column in [self.data_column, self.label_column]:
            if column not in training_data.columns:
                logger.warning(f"Column {column} not found in training data")
                return False
        return True

    def train(self, training_data: pd.DataFrame) -> None:
        """
        Train an intent-classifier on the provided training data.

        The training is skipped if the data-format is not valid.
        :param training_data: A pandas dataframe containing at least list of strings and corresponding labels.
        :type training_data: pd.DataFrame
        """
        if not self.validate(training_data):
            logger.warning(
                f"Training dataframe is invalid, for {self.__class__.__name__} plugin."
            )
            return

        if self.valid_mlpmodel:
            logger.warning(f"Model already exists on {self.mlp_model_path}")
            return

        skip_labels_filter = training_data[self.label_column].isin(self.skip_labels)
        training_data = training_data[~skip_labels_filter].copy()

        self.labels_num = training_data[self.label_column].nunique()
        sample_size = 5 if len(training_data) > 5 else len(training_data)
        training_data.rename(
            columns={self.data_column: const.TEXT, self.label_column: const.LABELS},
            inplace=True,
        )
        args = self.init_model()
        logger.debug(
            f"Displaying a few samples (this goes into the model):\n{training_data.sample(sample_size)}\nLabels: {self.labels_num}."
        )
        self.model_pipeline.fit(training_data[const.TEXT], training_data[const.LABELS])
        USE = "use"
        if args and args[const.USE_GRIDSEARCH][USE]:
            logger.debug(
                "Best gridsearch params found:\n"
                + str(
                    "\n".join(
                        str(items) for items in self.model_pipeline.best_params_.items()
                    )
                )
            )
        self.save()

    def save(self) -> None:
        """
        Save the plugin artifacts.

        :raises ValueError: In case the mlp model is not trained.
        """
        if not self.model_pipeline or not self.valid_mlpmodel:
            raise ValueError(
                f"Plugin {self.__class__.__name__} seems to be un-trained."
            )
        save_file(
            self.mlp_model_path,
            self.model_pipeline,
            mode="wb",
            writer=joblib.dump,
        )

    def load(self) -> None:
        """
        Load the plugin artifacts.
        """
        self.model_pipeline = load_file(
            self.mlp_model_path, mode="rb", loader=joblib.load
        )

    def utility(self, input: Input, _: Output) -> Any:
        return self.inference(
            input.clf_feature
        )  # pylint: disable=no-value-for-parameter



"""
.. _NumericalEntity:

Numerical Entity
============

Provides the entity class for representing numbers in natural language. This entity is obtained via :ref:`DucklingPlugin<DucklingPlugin>`.

Plugin Walkthrough
------------------

.. ipython::

    In [1]: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["number"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: duckling_plugin.parse("12")

    In [4]: duckling_plugin.parse("twenty two")

Workflow Integration
--------------------

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.workflow import Workflow

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["number"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: _, output = workflow.run(Input(utterances=["seventy two", "4 fruits"]))

    In [5]: output

CASTING
--------

We have noticed that there is a need to cast numerical values to time. The :ref:`as_time<number_entity_as_time>` method can be used to do this.
The default assumption is to use the number as :code:`day` but we can also replace the :code:`month` and :code:`hour`.

.. ipython::

    In [1]: import pytz
       ...: from datetime import datetime
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.base import Input
       ...: from dialogy.types import Intent
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.utils import dt2timestamp

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["number"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: reference_dt = datetime(2022, 1, 1, 12, 0, 0, tzinfo=pytz.timezone("Asia/Kolkata"))
       ...: reference_time = dt2timestamp(reference_dt)

    In [4]: entities = duckling_plugin.parse("12")

    In [5]: [entity.as_time(reference_time, "Asia/Kolkata") for entity in entities]

"""
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Optional

import attr

from dialogy import constants as const
from dialogy.types.entity.base_entity import BaseEntity
from dialogy.types.entity.time import TimeEntity
from dialogy.utils import unix_ts_to_datetime


@attr.s
class NumericalEntity(BaseEntity):
    """
    Numerical Entity Type

    Use this type for handling all entities that can be parsed to obtain:
    - numbers
    - date
    - time
    - datetime

    Attributes:
        - `dim` dimension of the entity from duckling parser
        - `type` is the type of the entity which can have values in ["value", "interval"]
    """

    origin = attr.ib(
        type=str, default="value", validator=attr.validators.instance_of(str)
    )
    entity_type: Optional[str] = attr.ib(default="number", order=False)

    @classmethod
    def from_duckling(
        cls, d: Dict[str, Any], alternative_index: int
    ) -> NumericalEntity:
        value = d[const.VALUE][const.VALUE]
        return cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            values=[{const.VALUE: value}],
            latent=d[const.LATENT],
        )

    def as_time(
        self, reference_unix_ts: int, timezone: str, replace: str = "day"
    ) -> TimeEntity:
        """
        Converts a duration entity to a time entity.

        .. _number_entity_as_time:
        """
        reference_datetime: datetime = unix_ts_to_datetime(
            reference_unix_ts, timezone=timezone
        )
        if replace not in [const.HOUR, const.DAY, const.MONTH]:
            raise RuntimeError(
                f"Expected replace to be one of {[const.HOUR, const.DAY, const.MONTH]}"
            )
        value = reference_datetime.replace(**{replace: self.value}).isoformat()

        entity = TimeEntity(
            range={
                const.START: self.range[const.START],
                const.END: self.range[const.END],
            },
            body=self.body,
            score=self.score,
            dim="time",
            alternative_index=self.alternative_index,
            latent=self.latent,
            values=[{const.VALUE: value}],
            grain=replace,
        )
        entity.set_entity_type()
        return entity



from typing import Any, Dict, List, Optional

import attr
from pydash import py_

from dialogy import constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.types import BaseEntity, TimeEntity


def has_time_component(entity: BaseEntity) -> bool:
    return entity.entity_type in [
        CombineDateTimeOverSlots.TIME,
        CombineDateTimeOverSlots.DATETIME,
    ]


def is_date(entity: BaseEntity) -> bool:
    return entity.entity_type == CombineDateTimeOverSlots.DATE


class CombineDateTimeOverSlots(Plugin):
    """

    .. _CombineDateTimeOverSlots:

    Dialog case:

    Assume that at the moment of dialog the date is 15th December 2019.

    +---+------+----------------------------+-------------------------------------------------------+
    |   |      |          utterance         |                         entity                        |
    +===+======+============================+=======================================================+
    | 1 | BOT  | When do you want to visit? |                                                       |
    +---+------+----------------------------+-------------------------------------------------------+
    | 2 | USER | Day after tomorrow         |        {'type': 'date', 'value': '2019-12-17'}        |
    +---+------+----------------------------+-------------------------------------------------------+
    | 3 | BOT  | At what time?              |                                                       |
    +---+------+----------------------------+-------------------------------------------------------+
    | 4 | USER | 3 pm                       | {'type': 'time', 'value': '2019-12-15T15:00:00+0000'} |
    +---+------+----------------------------+-------------------------------------------------------+

    We want to use the slots such that the date from interaction [2] and the time from interaction [4] are combined into a single entity.

    We receive this as a tracked slot:

    .. code-block:: json

        [{
            "name": "_callback_",
            "slots": [{
                "name": "callback_datetime",
                "type": [
                    "time",
                    "date",
                    "datetime"
                ],
                "values": [{
                    "alternative_index": 0,
                    "body": "tomorrow",
                    "entity_type": "date",
                    "grain": "day",
                    "parsers": ["duckling"],
                    "range": {
                        "end": 8,
                        "start": 0
                    },
                    "score": null,
                    "type": "value",
                    "value": "2021-10-15T00:00:00+05:30"
                }]
            }]
        }]

    At interaction [4] we would have a TimeEntity from DucklingPlugin but as we can see, the tracked slot provides us
    previously filled values as json.

    So we need to parse the first value in the slot (since they are the entities) if they are one of these date, time, datetime types.
    This would give us a date / time entity that we can subsequently combine with the available entity in the current turn.
    """

    DATE = "date"
    TIME = "time"
    DATETIME = "datetime"
    SUPPORTED_ENTITIES = [DATE, TIME, DATETIME]

    def __init__(
        self,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        replace_output: bool = True,
        use_transform: bool = False,
        trigger_intents: Optional[List[str]] = None,
        debug: bool = False,
    ) -> None:
        super().__init__(
            dest=dest,
            guards=guards,
            replace_output=replace_output,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
            debug=debug,
        )
        self.trigger_intents = trigger_intents

    def join(
        self, current_entity: TimeEntity, previous_entity: TimeEntity
    ) -> BaseEntity:
        current_turn_datetime = current_entity.get_value()
        previous_turn_datetime = previous_entity.get_value()

        if is_date(current_entity) and has_time_component(previous_entity):
            combined_value = current_turn_datetime.replace(
                hour=previous_turn_datetime.hour,
                minute=previous_turn_datetime.minute,
                second=previous_turn_datetime.second,
            )
        elif is_date(previous_entity) and has_time_component(current_entity):
            combined_value = current_turn_datetime.replace(
                year=previous_turn_datetime.year,
                month=previous_turn_datetime.month,
                day=previous_turn_datetime.day,
            )
        else:
            return current_entity
        return TimeEntity.from_dict(
            {const.VALUE: combined_value.isoformat()}, current_entity
        )

    def get_tracked_slots(
        self, slot_tracker: Optional[List[Dict[str, Any]]]
    ) -> List[Dict[str, Any]]:
        if not self.trigger_intents or not slot_tracker:
            return []

        tracked_intents = [
            intent
            for intent in slot_tracker
            if intent[const.NAME] in self.trigger_intents
        ]

        if not tracked_intents:
            return []

        tracked_intent, *_ = tracked_intents
        return tracked_intent.get(const.SLOTS, [])

    def pick_previously_filled_time_entity(
        self, tracked_slots: Optional[List[Dict[str, Any]]]
    ) -> Optional[TimeEntity]:
        if not tracked_slots:
            return None

        filled_entities_json = tracked_slots[0][const.VALUES]

        if not filled_entities_json or not isinstance(filled_entities_json, list):
            return None

        filled_entity_json, *_ = filled_entities_json
        filled_entity_json[const.VALUES] = [
            {const.VALUE: filled_entity_json[const.VALUE]}
        ]

        return TimeEntity(**filled_entity_json)

    def combine_time_entities_from_slots(
        self, slot_tracker: Optional[List[Dict[str, Any]]], entities: List[BaseEntity]
    ) -> List[BaseEntity]:
        """
        Combines the time entities from the slots with the entities from the current turn.

        .. _combinetimeentitiesfromslots:

        :param slot_tracker: The slot tracker from the previous turn.
        :type slot_tracker: Optional[List[Dict[str, Any]]]
        :param entities: Entities found in the current turn.
        :type entities: List[BaseEntity]
        :return: Combined set of entities.
        :rtype: List[BaseEntity]
        """
        previously_filled_time_entity = self.pick_previously_filled_time_entity(
            self.get_tracked_slots(slot_tracker)
        )

        if not previously_filled_time_entity:
            return entities

        time_entities, other_entities = py_.partition(
            entities,
            lambda entity: entity.entity_type
            in CombineDateTimeOverSlots.SUPPORTED_ENTITIES,
        )
        combined_time_entities = [
            self.join(entity, previously_filled_time_entity) for entity in time_entities
        ]
        return combined_time_entities + other_entities

    def utility(self, input: Input, output: Output) -> List[BaseEntity]:
        """
        Combine the date and time entities collected across turns into a single entity.
        """
        return self.combine_time_entities_from_slots(
            input.slot_tracker, output.entities
        )



from typing import Any, Dict, Optional

from dialogy import constants as const
from dialogy.types.entity.amount_of_money import CurrencyEntity
from dialogy.types.entity.base_entity import BaseEntity
from dialogy.types.entity.credit_card_number import CreditCardNumberEntity
from dialogy.types.entity.duration import DurationEntity
from dialogy.types.entity.numerical import NumericalEntity
from dialogy.types.entity.people import PeopleEntity
from dialogy.types.entity.time import TimeEntity
from dialogy.types.entity.time_interval import TimeIntervalEntity


def deserialize_duckling_entity(
    duckling_entity_dict: Dict[str, Any],
    alternative_index: int,
    reference_time: Optional[int] = None,
    timezone: str = "UTC",
    duration_cast_operator: Optional[str] = None,
) -> BaseEntity:
    keys = tuple(sorted(duckling_entity_dict.keys()))
    if keys != const.DUCKLING_ENTITY_KEYS:
        raise ValueError(
            f"Invalid Duckling entity keys: {keys} expected {const.DUCKLING_ENTITY_KEYS}."
        )

    dimension = duckling_entity_dict[const.DIM]
    if dimension not in const.DUCKLING_DIMS:
        raise ValueError(
            f"Invalid Duckling dimension: {dimension} expected {const.DUCKLING_DIMS}."
        )

    value_keys = tuple(sorted(duckling_entity_dict[const.VALUE].keys()))

    if dimension == const.PEOPLE:
        return PeopleEntity.from_duckling(duckling_entity_dict, alternative_index)

    elif dimension == const.TIME:
        if value_keys == const.DUCKLING_TIME_VALUES_ENTITY_KEYS:
            return TimeEntity.from_duckling(duckling_entity_dict, alternative_index)

        elif (
            dimension == const.TIME
            and value_keys in const.DUCKLING_TIME_INTERVAL_ENTITY_KEYS
        ):
            return TimeIntervalEntity.from_duckling(
                duckling_entity_dict, alternative_index
            )

        else:
            raise ValueError(
                f"Dimension {const.TIME} recieved invalid keys {value_keys} expected either {const.DUCKLING_TIME_VALUES_ENTITY_KEYS} or {const.DUCKLING_TIME_INTERVAL_ENTITY_KEYS}."
            )

    elif dimension == const.AMOUNT_OF_MONEY:
        return CurrencyEntity.from_duckling(duckling_entity_dict, alternative_index)

    elif dimension == const.NUMBER:
        return NumericalEntity.from_duckling(duckling_entity_dict, alternative_index)

    elif dimension == const.DURATION:
        entity = DurationEntity.from_duckling(duckling_entity_dict, alternative_index)
        if duration_cast_operator and reference_time:
            return entity.as_time(reference_time, timezone, duration_cast_operator)
        return entity
    else:  # dimension == const.CREDIT_CARD_NUMBER:
        return CreditCardNumberEntity.from_duckling(
            duckling_entity_dict, alternative_index
        )



"""
.. _Output:

The :ref:`Output <Output>` class creates immutable instances that describe the output of a single turn of a conversation.
The output contains 

#. :code:`List[Intent]` sorted in descending order of confidence scores.
#. :code:`List[BaseEntity]` in an arbitrary order.

We don't require the :code:`List[BaseEntity]` apart from logging purposes. We fill relevant entities within :ref:`slots <Slot>`
of the predicted :ref:`Intent<Intent>`. Currently this is done only for the :ref:`Intent<Intent>` with the highest confidence score.
That's because currently, we only process a single trigger from the SLU API (at the dialog manager), which is an :ref:`Intent<Intent>`.

.. admonition:: Why do I see :ref:`Input <Input>` and :ref:`Output <Output>` as inputs to all :ref:`Plugins <AbstractPlugin>`?

    It is a common pattern for all the plugins to require both as arguments. Since this could be confusing nomenclature, :ref:`Input <Input>`
    and :ref:`Output <Output>` bear meaning and even separation for the SLU API, **not for** :ref:`Plugins <AbstractPlugin>`.

Setting Output
--------------

Just like :ref:`Input <Input>`, :ref:`Output <Output>` is also an immutable class.

#. **Don't change the attribute elements in place.** It would work for now
   but even types like `Intent` and `BaseEntity` will soon become immutable as well.

#. Plugins that produce output must always produce a :code:`List[Intent]` or :code:`List[BaseEntity]`.
   Say, returning a single :code:`Intent` would break validations on the output class.


Serialization
-------------

If there is a need to represent an :ref:`Input<Input>` as a `dict` we can do the following:

.. ipython::

    In [1]: from dialogy.base import Output
       ...: from dialogy.types import Intent

    In [2]: output = Output(intents=[Intent(name="_greeting_", score=1.0)])

    In [3]: output.json()
"""
from __future__ import annotations

from typing import Any, Dict, List, Optional

import attr

from dialogy.types import BaseEntity, Intent


@attr.frozen
class Output:
    """
    Represents output of the SLU API.
    """

    intents: List[Intent] = attr.ib(default=attr.Factory(list), kw_only=True)
    """
    A list of intents. Produced by :ref:`XLMRMultiClass <XLMRMultiClass>`
    or :ref:`MLPMultiClass <MLPMultiClass>`.
    """

    entities: List[BaseEntity] = attr.ib(default=attr.Factory(list), kw_only=True)
    """
    A list of entities. Produced by :ref:`DucklingPlugin <DucklingPlugin>`, 
    :ref:`ListEntityPlugin <ListEntityPlugin>` or :ref:`ListSearchPlugin <ListSearchPlugin>`.
    """

    @intents.validator  # type: ignore
    def _are_intents_valid(
        self, _: attr.Attribute, intents: List[Intent]  # type: ignore
    ) -> None:
        if not isinstance(intents, list):
            raise TypeError(f"intents must be a list, not {type(intents)}")

        if not intents:
            return

        if any(not isinstance(intent, Intent) for intent in intents):
            raise TypeError(
                f"intents must be a List[Intent] but {intents} was provided."
            )

    @entities.validator  # type: ignore
    def _are_entities_valid(
        self, _: attr.Attribute, entities: List[BaseEntity]  # type: ignore
    ) -> None:
        if not isinstance(entities, list):
            raise TypeError(f"entities must be a list, not {type(entities)}")

        if not entities:
            return

        if any(not isinstance(entity, BaseEntity) for entity in entities):
            raise TypeError(
                f"intents must be a List[BaseEntity] but {entities} was provided."
            )

    def json(self: Output) -> Dict[str, List[Dict[str, Any]]]:
        """
        Serialize `Output`_ to a JSON-like dict.

        :param self: [description]
        :type self: Output
        :return: [description]
        :rtype: Dict[str, List[Dict[str, Any]]]
        """
        return {
            "intents": [intent.json() for intent in self.intents],
            "entities": [entity.json() for entity in self.entities],
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any], reference: Optional[Output] = None) -> Output:
        """
        Create a new `Output`_ instance from a dictionary.

        :param d: A dictionary such that keys are a subset of `Output`_ attributes.
        :type d: Dict[str, Any]
        :param reference: An existing `Output`_ instance., defaults to None
        :type reference: Optional[Output], optional
        :return: A new `Output`_ instance.
        :rtype: Output
        """
        if reference:
            return attr.evolve(reference, **d)
        return attr.evolve(cls(), **d)



import json
from typing import Any, Dict, List, Optional, Tuple

from pydash import py_

from dialogy.types import BaseEntity
from dialogy.utils import normalize


def entity_scoring(presence: int, input_size: int) -> float:
    return presence / input_size


class EntityScoringMixin:
    """
    Mixin for scoring and aggregation of entities over a set of transcripts.

    .. _EntityScoringMixin:
    """

    threshold: Optional[float] = None
    """
    Value to compare against an entity's score.
    """

    def remove_low_scoring_entities(
        self, entities: List[BaseEntity]
    ) -> List[BaseEntity]:
        """
        Remove entities with a lower score than the threshold.

        This doesn't apply to entities with score=None.

        :param entities: A list of entities.
        :type entities: List[BaseEntity]
        :return: A list of entities with score higher than configured threshold.
        :rtype: List[BaseEntity]
        """
        if self.threshold is None:
            return entities

        high_scoring_entities = []
        for entity in entities:
            if entity.score is None:
                high_scoring_entities.append(entity)

            if entity.score is not None and self.threshold < entity.score:
                high_scoring_entities.append(entity)

        return high_scoring_entities

    def aggregate_entities(
        self,
        entity_type_value_group: Dict[Tuple[str, Any], List[BaseEntity]],
        input_size: int,
    ) -> List[BaseEntity]:
        """
        Reduce entities sharing same type and value.

        - Entities with same type and value are considered identical even if other metadata is same.
          These entities are part of a group.
        - We track the transcript indices for every entity in a group.
        - Select the minimum of all the indices. (because 0th transcript has highest confidence)
        - We pick one entity per group and modify its index to the minimum and score is aggregated for the group.
        - The entity picked is added to a list of aggregates.

        The above is done for all entities in a group

        :param entity_type_val_group: A data-structure that groups entities by type and value.
        :type entity_type_val_group: Dict[Tuple[str, Any], List[BaseEntity]]
        :return: A list of de-duplicated entities.
        :rtype: List[BaseEntity]
        """
        aggregated_entities = []
        for entities in entity_type_value_group.values():
            indices = [
                entity.alternative_index
                for entity in entities
                if isinstance(entity.alternative_index, int)
            ]
            min_alternative_index = py_.min_(indices) if indices else None
            representative = entities[0]
            representative.alternative_index = min_alternative_index
            representative.alternative_indices = indices
            representative.score = entity_scoring(len(py_.uniq(indices)), input_size)
            aggregated_entities.append(representative)
        return aggregated_entities

    def apply_filters(self, entities: List[BaseEntity]) -> List[BaseEntity]:
        """
        Filter entities with score less than the threshold.

        :param entities: A list of entities.
        :type entities: List[BaseEntity]
        :return: A list of entities. This can be at most the same length as `entities`.
        :rtype: List[BaseEntity]
        """
        return self.remove_low_scoring_entities(entities)

    def entity_consensus(
        self, entities: List[BaseEntity], input_size: int
    ) -> List[BaseEntity]:
        """
        Combine entities by type and value.

        This issue:
        https://github.com/Vernacular-ai/dialogy/issues/52
        Points at the problems where we can return multiple identical entities,
        depending on the number of transcripts that contain same body.

        :param entities: A list of entities which may have duplicates.
        :type entities: List[BaseEntity]
        :return: A list of entities scored and unique by type and value.
        :rtype: List[BaseEntity]
        """
        entity_type_value_group = py_.group_by(
            entities, lambda entity: (entity.type, entity.get_value())
        )
        aggregate_entities = self.aggregate_entities(
            entity_type_value_group, input_size
        )
        return self.apply_filters(aggregate_entities)

    @staticmethod
    def make_transform_values(transcript: Any) -> List[str]:
        """
        Make transcripts from a string/json-string.

        :param transcript: A string to search entities within.
        :type transcript: str
        :return: List of transcripts.
        :rtype: List[str]
        """
        try:
            transcript = json.loads(transcript)
            return normalize(transcript)
        except (json.JSONDecodeError, TypeError):
            return normalize(transcript)



"""
Module provides access to logger.

This needs to be used sparingly, prefer to raise specific exceptions instead.
"""
import sys

from loguru import logger

config = {
    "handlers": [
        {
            "sink": sys.stdout,
            "format": """
-------------------------------------------------------
<level>{level}</level>
-------
TIME: <green>{time}</green>
FILE: {name}:L{line} <blue>{function}(...)</blue>
<level>{message}</level>
-------------------------------------------------------
""",
            "colorize": True,
        },
        {
            "sink": "file.log",
            "rotation": "500MB",
            "retention": "10 days",
            "format": "{time} {level} -\n{message}\n--------------------\n",
        },
    ]
}

logger.configure(**config)
logger.enable(__name__)



"""
.. _xlmr_classifier:

This module provides a trainable XLMR classifier.
[read-more](https://arxiv.org/abs/1911.02116)
"""
import importlib
import os
import pickle
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
from sklearn import preprocessing

import dialogy.constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.types import Intent
from dialogy.utils import load_file, logger, save_file


class XLMRMultiClass(Plugin):
    """
    This plugin provides a classifier based on `XLM-Roberta <https://arxiv.org/abs/1911.02116>`.

    .. _XLMRMultiClass:
    The use_state flag in the XLMRMultiClass plugin is used to enable the use of state variable as the part of the text input.
    """

    def __init__(
        self,
        model_dir: str,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        debug: bool = False,
        threshold: float = 0.1,
        use_cuda: bool = False,
        score_round_off: int = 5,
        purpose: str = const.TRAIN,
        fallback_label: str = const.ERROR_LABEL,
        use_state: bool = False,
        data_column: str = const.DATA,
        label_column: str = const.LABELS,
        state_column: str = const.STATE,
        args_map: Optional[Dict[str, Any]] = None,
        skip_labels: Optional[List[str]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        try:
            classifer = getattr(
                importlib.import_module(const.XLMR_MODULE), const.XLMR_MULTI_CLASS_MODEL
            )
        except ModuleNotFoundError as error:
            raise ModuleNotFoundError(
                "Plugin requires simpletransformers -- https://simpletransformers.ai/docs/installation/"
            ) from error

        super().__init__(dest=dest, guards=guards, debug=debug)
        self.labelencoder = preprocessing.LabelEncoder()
        self.classifier = classifer
        self.model: Any = None
        self.model_dir = model_dir
        self.fallback_label = fallback_label
        self.data_column = data_column
        self.label_column = label_column
        self.state_column = state_column
        self.use_cuda = use_cuda
        self.use_state = use_state
        self.labelencoder_file_path = os.path.join(
            self.model_dir, const.LABELENCODER_FILE
        )
        self.threshold = threshold
        self.skip_labels = set(skip_labels or set())
        self.purpose = purpose
        self.round = score_round_off
        if args_map and (
            const.TRAIN not in args_map
            or const.TEST not in args_map
            or const.PRODUCTION not in args_map
        ):
            raise ValueError(
                f"Attempting to set invalid {args_map=}. "
                "It is missing some of {const.TRAIN}, {const.TEST}, {const.PRODUCTION} in configs."
            )
        self.args_map = args_map
        self.kwargs = kwargs or {}
        try:
            if os.path.exists(self.labelencoder_file_path):
                self.init_model()
        except EOFError:
            logger.error(
                f"Plugin {self.__class__.__name__} Failed to load labelencoder from {self.labelencoder_file_path}. "
                "Ignore this message if you are training but if you are using this in production or testing, then this is serious!"
            )

    def init_model(self, label_count: Optional[int] = None) -> None:
        """
        Initialize the model if artifacts are available.

        :param label_count: number of labels to train on or predict, defaults to None
        :type label_count: Optional[int], optional
        :raises ValueError: In case n is not provided or can't be calculated.
        """
        model_dir = const.XLMR_MODEL_TIER
        if os.path.exists(self.labelencoder_file_path):
            self.load()
            label_count = len(self.labelencoder.classes_)
            model_dir = self.model_dir
        if not label_count:
            raise ValueError(
                f"Plugin {self.__class__.__name__} needs either the training data or an existing labelencoder to initialize."
            )
        args = (
            self.args_map[self.purpose]
            if self.args_map and self.purpose in self.args_map
            else {}
        )
        try:
            self.model = self.classifier(
                const.XLMR_MODEL,
                model_dir,
                num_labels=label_count,
                use_cuda=self.use_cuda,
                args=args,
                **self.kwargs,
            )
        except OSError:
            self.model = self.classifier(
                const.XLMR_MODEL,
                const.XLMR_MODEL_TIER,
                num_labels=label_count,
                use_cuda=self.use_cuda,
                args=args,
                **self.kwargs,
            )

    @property
    def valid_labelencoder(self) -> bool:
        return hasattr(self.labelencoder, "classes_")

    def inference(self, texts: Optional[List[str]]) -> List[Intent]:
        """
        Predict the intent of a list of texts.
        If the model has been trained using the state features, it expects the text to also be appended with the state token else the predictions would be spurious.

        :param texts: A list of strings, derived from ASR transcripts.
        :type texts: List[str]
        :raises AttributeError: In case the labelencoder is not available.
        :return: A list of intents corresponding to texts.
        :rtype: List[Intent]
        """
        logger.debug(f"Classifier input:\n{texts}")
        fallback_output = Intent(name=self.fallback_label, score=1.0).add_parser(self)
        if not texts:
            logger.error(f"texts passed to model {texts}!")
            return [fallback_output]

        if self.model is None:
            logger.error(f"No model found for plugin {self.__class__.__name__}!")
            return [fallback_output]
        if self.use_state:
            for text in texts:
                if "<st>" not in text:
                    logger.error(
                        f"use_state = True but {text} doesn't contain <st> </st> tags, \
                        which means that training data is different from this data. This can lead to anomalous results"
                    )
                    return [fallback_output]
        if not self.valid_labelencoder:
            raise AttributeError(
                "Seems like you forgot to "
                f"save the {self.__class__.__name__} plugin."
            )

        predictions, logits = self.model.predict(texts)
        if not predictions:
            return [fallback_output]

        confidence_scores = [np.exp(logit) / sum(np.exp(logit)) for logit in logits]
        intents_confidence_order = np.argsort(confidence_scores)[0][::-1]
        predicted_intents = self.labelencoder.inverse_transform(
            intents_confidence_order
        )
        ordered_confidence_scores = [
            confidence_scores[0][idx] for idx in intents_confidence_order
        ]

        return [
            Intent(name=intent, score=round(score, self.round)).add_parser(
                self.__class__.__name__
            )
            for intent, score in zip(predicted_intents, ordered_confidence_scores)
        ]

    def validate(self, training_data: pd.DataFrame) -> bool:
        """
        Validate the training data is in the appropriate format

        :param training_data: A pandas dataframe containing at least list of strings and corresponding labels.
            Should also contain a state key if use_state = True while initializing object.
        :type training_data: pd.DataFrame
        :return: True if the dataframe is valid, False otherwise.
        :rtype: bool
        """
        if training_data.empty:
            logger.error("Training dataframe is empty.")
            return False
        expected_columns = [self.data_column, self.label_column]
        if self.use_state:
            expected_columns.append(self.state_column)
        for column in expected_columns:
            if column not in training_data.columns:
                logger.warning(f"Column {column} not found in training data")
                return False
        return True

    def train(self, training_data: pd.DataFrame) -> None:
        """
        Train an intent-classifier on the provided training data.

        The training is skipped if the data-format is not valid.
        While training with the use_state flag as true, make sure that the state column is the part of the training_data dataframe
        :param training_data: A pandas dataframe containing at least list of strings and corresponding labels.
        :type training_data: pd.DataFrame
        """
        if not self.validate(training_data):
            logger.warning(
                f"Training dataframe is invalid, for {self.__class__.__name__} plugin."
            )
            return

        skip_labels_filter = training_data[self.label_column].isin(self.skip_labels)
        training_data = training_data[~skip_labels_filter].copy()

        encoder = self.labelencoder.fit(training_data[self.label_column])
        sample_size = 5 if len(training_data) > 5 else len(training_data)
        training_data.rename(
            columns={self.data_column: const.TEXT, self.label_column: const.LABELS},
            inplace=True,
        )
        training_data.loc[:, const.LABELS] = encoder.transform(
            training_data[const.LABELS]
        )
        # Add state as an additonal field to text
        if self.use_state:
            training_data[const.TEXT] += "<st> " + training_data["state"] + " </st>"
        training_data = training_data[[const.TEXT, const.LABELS]]
        self.init_model(len(encoder.classes_))
        logger.debug(
            f"Displaying a few samples (this goes into the model):\n{training_data.sample(sample_size)}\nLabels: {len(encoder.classes_)}."
        )
        self.model.train_model(training_data)
        self.save()

    def save(self) -> None:
        """
        Save the plugin artifacts.

        :raises ValueError: In case the labelencoder is not trained.
        """
        if not self.model or not self.valid_labelencoder:
            raise ValueError(
                f"Plugin {self.__class__.__name__} seems to be un-trained."
            )
        save_file(
            self.labelencoder_file_path,
            self.labelencoder,
            mode="wb",
            writer=pickle.dump,
        )

    def load(self) -> None:
        """
        Load the plugin artifacts.
        """
        self.labelencoder = load_file(
            self.labelencoder_file_path, mode="rb", loader=pickle.load
        )

    def utility(self, input: Input, _: Output) -> List[Intent]:
        return self.inference(input.clf_feature)



"""
.. _slot_filler:

To handle a single dialog we need a system that can:

1. Classify inputs into categories.
2. Extract tokens of additional significance.
3. Check for associations between the two.

A slot-filler does the third bit. Once an :ref:`Intent <intent>` and a set of
:ref:`Entity <base_entity>` are identified, there needs to be an arrangement that
tells us if any of these are related to each other.

As per the current design, an :ref:`Intent <intent>` can contain a number of :ref:`Slot <slot>`. A :ref:`Slot <slot>`
fills an entity of a pre-defined type.

- A :ref:`Slot <slot>` configured to fill a :ref:`TimeEntity <time_entity>` will not fill a :ref:`KeywordEntity <KeywordEntity>`.
- A :ref:`Slot <slot>` will not fill more than one entity even if the type matches, unless :code:`fill_multiple` is provided.
"""



"""
Type definition for Slots. Read :ref:`Intent<Intent>` and :ref:`Slot Filling<Slot Filling>` for details.
"""
from __future__ import annotations

from typing import Any, Dict, List

import attr

from dialogy.types import BaseEntity


@attr.s
class Slot:
    """
    Slot Type

    .. _Slot:

    Keys are:
    - `name` of the slot
    - `type` is a list of types that can fill this slot
    - `values` list of entities extracted
    """

    name: str = attr.ib(kw_only=True, order=True)
    types: List[str] = attr.ib(kw_only=True, factory=list, order=False)
    values: List[BaseEntity] = attr.ib(kw_only=True, factory=list, order=False)

    def add(self, entity: BaseEntity) -> Slot:
        """
        Insert the `BaseEntity` within the current `Slot` instance.
        """
        self.values.append(entity)
        return self

    def clear(self) -> Slot:
        """
        Remove all `BaseEntity` within the current `Slot` instance.
        """
        self.values = []
        return self

    def json(self) -> Dict[str, Any]:
        """
        Convert the object to a dictionary.

        Returns:
            Dict[str, Any]
        """
        return {
            "name": self.name,
            "types": self.types,
            "values": [entity.json() for entity in self.values],
        }


Rule = Dict[str, Dict[str, Any]]



from dialogy.plugins.text.calibration.xgb import CalibrationModel
from dialogy.plugins.text.canonicalization import CanonicalizationPlugin
from dialogy.plugins.text.classification.mlp import MLPMultiClass
from dialogy.plugins.text.classification.xlmr import XLMRMultiClass
from dialogy.plugins.text.combine_date_time import CombineDateTimeOverSlots
from dialogy.plugins.text.duckling_plugin import DucklingPlugin
from dialogy.plugins.text.lb_plugin import DucklingPluginLB
from dialogy.plugins.text.list_entity_plugin import ListEntityPlugin
from dialogy.plugins.text.list_search_plugin import ListSearchPlugin
from dialogy.plugins.text.merge_asr_output import MergeASROutputPlugin
from dialogy.plugins.text.slot_filler.rule_slot_filler import RuleBasedSlotFillerPlugin



"""
.. _TimeIntervalEntity:

Time Interval Entity
====================

Provides the entity class for representing time intervals in natural language. This entity is obtained via :ref:`DucklingPlugin<DucklingPlugin>`.

Plugin Walkthrough
------------------

.. ipython::

    In [1]: from dialogy.plugins import DucklingPlugin

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: duckling_plugin.parse("weekend")

    In [4]: duckling_plugin.parse("between 4 to 5pm")

Workflow Integration
--------------------

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.plugins import DucklingPlugin
       ...: from dialogy.workflow import Workflow

    In [2]: duckling_plugin = DucklingPlugin(
       ...:     dest="output.entities",
       ...:     dimensions=["time"],
       ...:     locale="en_IN",
       ...:     timezone="Asia/Kolkata",
       ...: )

    In [3]: workflow = Workflow([duckling_plugin])

    In [4]: _, output = workflow.run(Input(utterances=["weekend", "between 4 to 5pm"]))

    In [5]: output
"""
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional

import attr

from dialogy import constants as const
from dialogy.types.entity.time import TimeEntity


@attr.s
class TimeIntervalEntity(TimeEntity):
    """
    Entities that can be parsed to obtain date, time or datetime interval.

    - "I need a flight between 6 am to 10 am."
    - "I have a flight at 6 am to 5 pm today."

    Attributes:
    - `origin`
    - `value` is a Dictionary which has either keys 'from' and 'to' or both
    """

    origin = "interval"
    dim = "time"
    value_keys = {const.FROM, const.TO, const.TYPE}
    type: str = attr.ib(default="value", order=False)
    from_value: Optional[datetime] = attr.ib(default=None, order=False)
    to_value: Optional[datetime] = attr.ib(default=None, order=False)
    values: List[Dict[str, Any]] = attr.ib(default=None, kw_only=True)
    value: Dict[str, Any] = attr.ib(default=None, kw_only=True)

    @values.validator  # type: ignore
    def _check_values(
        self, attribute: attr.Attribute, values: List[Dict[str, Any]]  # type: ignore
    ) -> None:
        if not values:
            return
        for value in values:
            obj_keys = set(value.keys())
            if not obj_keys.issubset(self.value_keys):
                raise TypeError(
                    f"Expected {obj_keys} to be a subset of {self.value_keys} for {attribute.name}"
                )

    def __attrs_post_init__(self) -> None:
        if self.values and not self.value:
            self.from_value = self.values[0].get(const.FROM, {}).get(const.VALUE)
            self.to_value = self.values[0].get(const.TO, {}).get(const.VALUE)
            self.value = {const.FROM: self.from_value, const.TO: self.to_value}
        elif self.value and not self.values:
            self.from_value = self.value.get(const.FROM, {})
            self.to_value = self.value.get(const.TO, {})
            self.values = [
                {
                    const.FROM: {const.VALUE: self.from_value},
                    const.TO: {const.VALUE: self.to_value},
                }
            ]

    def collect_datetime_values(self) -> List[datetime]:
        """
        Collect all datetime values from the entity

        :return: List of datetime values
        :rtype: List[str]
        """
        datetime_values = []
        for value in self.values:
            from_value = value.get(const.FROM, {}).get(const.VALUE)
            to_value = value.get(const.TO, {}).get(const.VALUE)
            datetime_values.append(datetime.fromisoformat(from_value or to_value))
        return datetime_values

    def get_value(self) -> Any:
        """
        Return the date string in ISO format from the dictionary passed

        .. code-block:: python

            date = {
                "from": {
                    "value": "2021-04-16T16:00:00.000+05:30",
                    "grain": "hour"
                },
                "type": "interval"
            }

        :param date: Dictionary which stores the datetime in ISO format, grain and type
        :type date: Dict[str, str]
        :return: :code:`date["value"]`
        :rtype: Optional[datetime]
        """
        value = self.value.get(const.FROM) or self.value.get(const.TO)
        if value:
            return datetime.fromisoformat(value)
        else:
            raise KeyError(
                f"Expected at least 1 of `from` or `to` in {self.values} for {self}"
            )

    @classmethod
    def from_duckling(cls, d: Dict[str, Any], alternative_index: int) -> TimeEntity:
        from_value = d[const.VALUE].get(const.FROM)
        to_value = d[const.VALUE].get(const.TO)
        grain_source = from_value or to_value
        grain = grain_source.get(const.GRAIN)
        time_interval_entity = cls(
            range={const.START: d[const.START], const.END: d[const.END]},
            body=d[const.BODY],
            dim=d[const.DIM],
            alternative_index=alternative_index,
            latent=d[const.LATENT],
            values=d[const.VALUE][const.VALUES],
            type=d[const.VALUE][const.TYPE],
            grain=grain,
        )
        time_interval_entity.set_entity_type()
        return time_interval_entity
